# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import copy
import logging
import operator
import signal
import struct
import sys
import time
from datetime import datetime
from types import ModuleType
from typing import Generic, TypeVar

import numpy as np
import serial.tools.list_ports

from acconeer.exptool._core.communication.comm_devices import (
    get_serial_devices,
    tag_serial_ports_objects,
)


try:
    from PySide6 import QtCore
except ImportError:
    QtCore = None

try:
    import pyqtgraph as pg
except ImportError:
    pg = None

S = TypeVar("S")
T = TypeVar("T")
DTypeT = TypeVar("DTypeT")


class ExampleInterruptHandler:
    def __init__(self):
        self._signal_count = 0
        signal.signal(signal.SIGINT, self.interrupt_handler)

    @property
    def got_signal(self):
        return self._signal_count > 0

    def force_signal_interrupt(self):
        self.interrupt_handler(signal.SIGINT, None)

    def interrupt_handler(self, signum, frame):
        self._signal_count += 1
        if self._signal_count >= 3:
            raise KeyboardInterrupt


def mpl_setup_yaxis_for_phase(ax):
    ax.set_ylim(-np.pi, np.pi)
    ax.set_yticks(np.linspace(-np.pi, np.pi, 5))
    ax.set_yticklabels([r"$-\pi$", r"$-\pi/2$", r"0", r"$\pi/2$", r"$\pi$"])


def timestamp():
    """Get the current date and time formatted as 1999-12-31_23-59-59"""

    return datetime.now().strftime("%Y-%m-%d_%H-%M-%S")


def config_logging(args=None, level=logging.WARNING):
    fmt = "{asctime}.{msecs:03.0f} | {levelname:<7} | {processName:<16} | {name:<36} | {message}"
    datefmt = "%H:%M:%S"

    if args is not None:
        if args.debug:
            level = logging.DEBUG
        elif args.verbose:
            level = logging.INFO
        elif args.quiet:
            level = logging.ERROR

    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt, datefmt=datefmt, style="{")
    stream_handler.setFormatter(formatter)
    log = logging.getLogger(__name__.split(".")[0])
    log.setLevel(level)
    log.addHandler(stream_handler)

    logging.getLogger(__name__).debug("logging configured")


def set_loglevel(level):
    log = logging.getLogger(__name__.split(".")[0])
    log.setLevel(level)


def tag_serial_ports(port_infos):
    """
    Returns every port and Acconeer model number in a tuple if it's an
    Acconeer product.

    E.g:
    [ ("/dev/ttyUSB0", None),   <- Not an Acconeer product
      ("/dev/ttyUSB1", "XB112") <- An Acconeer product
    ]

    :param port_infos: Information about ports as given by
        `serial.tools.list_ports.comports()`
    :returns: List of tuples [(<port>, <model number> or None), ...]
    """
    port_tag_tuples = tag_serial_ports_objects(port_infos)
    port_tag_tuples = [(port.device, tag) for (port, tag) in port_tag_tuples]
    return port_tag_tuples


def get_tagged_serial_ports():
    tagged_serial_ports = []
    serial_devices = get_serial_devices()
    for serial_device in serial_devices:
        tagged_serial_ports.append((serial_device.port, serial_device.name))
    return tagged_serial_ports


def autodetect_serial_port():
    port_infos = serial.tools.list_ports.comports()

    tagged_serial_ports = tag_serial_ports(port_infos)
    acconeer_port_infos = [pinfo for pinfo in tagged_serial_ports if pinfo[1]]

    # Check for mutliple Acconeer devices;
    if len(acconeer_port_infos) > 1:
        print("Found multiple Acconeer products:", end="\n\n")
        for port, tag in [("Serial port:", "Model:")] + acconeer_port_infos:
            print(f"\t{port:<15} {tag:}")
        print('\nRun the script again and specify port using the "-u"/"--uart" flag.')
        sys.exit()

    if len(acconeer_port_infos) == 1:
        port, tag = acconeer_port_infos[0]
        print("Autodetected {} on {}\n".format(tag, port))
        return port

    for port_info in port_infos:
        port, desc, _ = port_info
        if desc == "FT230X Basic UART":
            print("Autodetected FT230X Basic UART on {}\n".format(port))
            return port

    if len(port_infos) == 0:
        print("Could not autodetect serial port, no ports available")
        sys.exit()
    elif len(port_infos) == 1:
        print("Autodetected single available serial port on {}\n".format(port))
        return port_infos[0][0]
    else:
        print("Multiple serial ports are available:")
        for port_info in port_infos:
            port, desc, _ = port_info
            print("  {:<13}  ({})".format(port, desc))
        print("\nRe-run the script with a given port")
        sys.exit()

    print("Could not autodetect serial port")
    sys.exit()


def get_module_version(module: ModuleType) -> str:
    if hasattr(module, "__version__"):
        return str(module.__version__)
    else:
        msg = f"Unknown module version for {module.__name__}"
        raise AttributeError(msg)


def color_cycler(i=0):
    category12 = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
        "#e6ab02",
        "#f516bd",
    ]

    return category12[i % len(category12)]


def pg_pen_cycler(i=0, style=None, width=2):
    pen = pg.mkPen(color_cycler(i), width=width)
    if style == "--":
        pen.setStyle(QtCore.Qt.DashLine)
    elif style is not None:
        pen.setStyle(style)
    return pen


def pg_brush_cycler(i=0):
    return pg.mkBrush(color_cycler(i))


class SmoothMax:
    def __init__(self, f=None, hysteresis=0.5, tau_decay=2.0, tau_grow=0.5):
        self.fixed_dt = 1 / f if (f is not None and f > 0) else None
        self.hyst = hysteresis
        self.tc_decay = tau_decay
        self.tc_grow = tau_grow

        self.last_t = 0
        self.x = self.y = -1

    def update(self, data):
        m = max(np.nanmax(data), 1e-12)

        if self.fixed_dt is None:
            now = time.perf_counter()
            dt = now - self.last_t
            self.last_t = now
        else:
            dt = self.fixed_dt

        ax = np.exp(-dt / self.tc_decay) if self.tc_decay > 1e-3 else 0
        ay = np.exp(-dt / self.tc_grow) if self.tc_grow > 1e-3 else 0

        if m > self.x:
            self.x = m
        else:
            self.x = (1 - ax) * m + ax * self.x

        if self.y < 0:
            self.y = self.x
        elif self.x > self.y:
            self.y = (1 - ay) * self.x * (1 + self.hyst) + ay * self.y
        elif self.x < (1 - self.hyst) * self.y:
            self.y = self.x / (1 - self.hyst)

        return self.y


class SmoothLimits:
    def __init__(self, f=None, hysteresis=0.3, tau_decay=1.5, tau_grow=0.3):
        self.fixed_dt = 1 / f if (f is not None and f > 0) else None
        self.hyst = hysteresis
        self.tc_decay = tau_decay
        self.tc_grow = tau_grow

        self.last_t = 0
        self.x = self.y = self.z = None

    def update(self, data):
        data_lims = (np.nanmin(data), np.nanmax(data))

        if self.fixed_dt is None:
            now = time.perf_counter()
            dt = now - self.last_t
            self.last_t = now
        else:
            dt = self.fixed_dt

        if self.x is None:  # First call
            self.x = list(data_lims)
            self.y = list(data_lims)
            self.z = list(data_lims)
            return list(data_lims)

        ad = np.exp(-dt / self.tc_decay) if self.tc_decay > 1e-3 else 0
        ag = np.exp(-dt / self.tc_grow) if self.tc_grow > 1e-3 else 0

        ops = (operator.lt, operator.gt)
        idxs = (0, 1)

        for op, i in zip(ops, idxs):
            if op(data_lims[i], self.x[i]):
                self.x[i] = data_lims[i]
            else:
                self.x[i] = (1 - ad) * data_lims[i] + ad * self.x[i]

        for op, i in zip(ops, idxs):
            if op(self.x[i], self.y[i]):
                self.y[i] = (1 - ag) * self.x[i] + ag * self.y[i]
            else:
                self.y[i] = self.x[i]

        for op, i, j in zip(ops, idxs, reversed(idxs)):
            y_hyst = self.y[i] + (self.y[i] - self.y[j]) * (1 / (1 - self.hyst) - 1)

            if op(self.y[i], self.z[i]):
                self.z[i] = self.y[i]
            elif op(self.z[i], y_hyst):
                self.z[i] = y_hyst

        return self.z


pg_phase_ticks = [
    list(zip(np.linspace(-np.pi, np.pi, 5), ["-π", "-π/2", "0", "π/2", "π"])),
    [(x, "") for x in np.linspace(-np.pi, np.pi, 9)],
]


def pg_setup_polar_plot(plot, max_r=1):
    plot.showAxis("left", False)
    plot.showAxis("bottom", False)
    plot.setAspectLocked()
    plot.disableAutoRange()
    s = 1.15
    plot.setXRange(-s * max_r, s * max_r)
    plot.setYRange(-s * max_r, s * max_r)

    for i, r in enumerate(np.linspace(0, max_r, 5)[1:]):
        circle = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
        circle.setPen(pg.mkPen("k" if i == 3 else 0.5))
        plot.addItem(circle)

        if i == 3:
            text_item = pg.TextItem(text=str(r), color="k", anchor=(0, 1))
            text_item.setPos(np.cos(np.pi / 8) * r, np.sin(np.pi / 8) * r)
            plot.addItem(text_item)

    for i in range(8):
        deg = (360 / 8) * i
        rad = np.radians(deg)
        x = np.cos(rad)
        y = np.sin(rad)
        text = str(int(deg)) + "\u00b0"
        ax = (-x + 1) / 2
        ay = (y + 1) / 2
        text_item = pg.TextItem(text, color="k", anchor=(ax, ay))
        text_item.setPos(max_r * x * 1.02, max_r * y * 1.02)
        plot.addItem(text_item)
        plot.plot([0, max_r * x], [0, max_r * y], pen=pg.mkPen(0.5))


def pg_mpl_cmap(name):
    import matplotlib.pyplot as plt

    cmap = plt.get_cmap(name)
    cmap._init()
    return (np.array(cmap._lut) * 255).astype(np.uint8)


class FreqCounter:
    """
    Helper class for measuring & calculating temporal information:
        * time between ticks (lp-filtered), `FreqCounter.lp_dt`
        * frequency
        * data throughput (if num_bits is not None)
    """

    def __init__(self, a=None, tc=None, num_bits=None):
        assert (a is None) or (tc is None)

        if (a is None) and (tc is None):
            tc = 1.0

        self.a = a
        self.tc = tc
        self.num_bits = num_bits
        self.avg_dt_buf_len = 150  # Chosen after empirical testing
        self.reset()

    def reset(self):
        self.last_t = None
        self.lp_avg_dt = None
        self.num_ticks = 0
        self.avg_dt_buf = np.full(self.avg_dt_buf_len, np.nan)

    def tick_values(self):
        """
        Ticks the FreqCounter by taking the time delta from last time this was called.
        The update is done by a running average over the last 5 sweeps.
        The averaged value is passed through a lp_filter to avoid displaying blinking numbers.

        OBS! The first call to this method will return None.

        :returns:
            3-tuple of (delta time (averaged and lp-filtered), frequency, [throughput])
            -- OR --
            None
        """

        now = time.perf_counter()

        if self.last_t is None:
            self.last_t = now
            return None

        dt = now - self.last_t

        self.avg_dt_buf = np.roll(self.avg_dt_buf, -1)
        self.avg_dt_buf[-1] = dt

        avg_dt = np.nanmean(self.avg_dt_buf)

        if self.a is not None:
            a = self.a
        else:
            a = np.exp(-avg_dt / self.tc)

        a = min(a, 1.0 - 1.0 / (1.0 + self.num_ticks))

        if self.lp_avg_dt is None:
            self.lp_avg_dt = avg_dt
        else:
            self.lp_avg_dt = a * self.lp_avg_dt + (1 - a) * avg_dt

        self.last_t = now
        self.num_ticks += 1

        f = 1 / self.lp_avg_dt

        if self.num_bits is None:
            data_rate = None
        else:
            data_rate = self.num_bits * f

        return self.lp_avg_dt, f, data_rate

    def tick(self):
        """
        Prints the values returned by `tick_values()`.
        """
        tick_info = self.tick_values()
        if tick_info is None:
            return

        dt, f, data_rate = tick_info
        dt_ms = dt * 1e3
        data_rate_mbps = data_rate * 1e-6

        if data_rate is None:
            print(" {:5.1f} ms, {:5.1f} Hz".format(dt_ms, f), end="\r")
        else:
            s = " {:5.1f} ms, {:5.1f} Hz, {:5.2f} Mbit/s".format(dt_ms, f, data_rate_mbps)
            print(s, end="\r")


def hex_to_rgb_tuple(hex_color):
    return struct.unpack("BBB", bytes.fromhex(hex_color.lstrip("#")))


def is_power_of_2(n):
    return (n & (n - 1) == 0) and n != 0


def optional_or_else(value, default):
    return default if value is None else value


class PhonySeries(Generic[T]):
    def __init__(self, prototype: T, is_prototype_singleton: bool = True) -> None:
        self._prototype = prototype
        self._is_prototype_singleton = is_prototype_singleton

    def __next__(self) -> T:
        if self._is_prototype_singleton:
            return self._prototype
        else:
            return copy.copy(self._prototype)

    def __iter__(self) -> PhonySeries:
        return self



# Copyright (c) Acconeer AB, 2022-2026
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Any, Callable, Dict, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtWidgets import QLabel, QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121, opser
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import distance
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.a121.algo.distance import (
    DetailedStatus,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
)
from acconeer.exptool.a121.algo.distance._detector import _load_algo_data
from acconeer.exptool.a121.algo.distance._detector_plugin import ViewPlugin as DistanceViewPlugin
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    GeneralMessage,
    GroupBox,
    HandledException,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    TwoSensorIdsEditor,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)

from ._configs import get_default_detector_config
from ._processor import Processor, ProcessorConfig, ProcessorResult


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_ids: list[int] = attrs.field(factory=lambda: [1, 1])
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    bilateration_config: ProcessorConfig = attrs.field(factory=ProcessorConfig)
    context: DetectorContext = attrs.field(factory=DetectorContext)
    replaying: bool = attrs.field(default=False)


def serialized_attrs_instance_has_diverged(attrs_instance: Any) -> bool:
    """Checks (recursively) if a de-serialized attrs-instances contains
    all attributes defined its respective class definition.

    :param attrs_config:
        An instance of an attrs-class, that possibly contains other attrs-instances.
    :returns: True if any attrs-instance have diverged from its class, False otherwise
    """
    # TODO: Should end up in the `Config` ABC
    attrs_class = type(attrs_instance)
    if not attrs.has(attrs_class):
        msg = f"Cannot check object of type {attrs_class!r}. It's not an attrs-class."
        raise TypeError(msg)

    for attribute in attrs_instance.__attrs_attrs__:
        try:
            value = getattr(attrs_instance, attribute.name)

            if attrs.has(type(value)) and serialized_attrs_instance_has_diverged(value):
                return True
        except AttributeError:
            log.info(
                f"Serialized object of type {attrs_class.__name__!r} "
                + "seems to have converged from its definition."
            )
            log.info(f"Should contain the attribute {attribute.name!r} but did not.")
            return True

    return False


class PluginPresetId(Enum):
    DEFAULT = auto()


@attrs.mutable(kw_only=True)
class BilaterationPreset:
    detector_config: DetectorConfig = attrs.field()
    bilateration_config: ProcessorConfig = attrs.field()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    bilateration_config: ProcessorConfig
    num_curves: int
    detector_config: DetectorConfig
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, BilaterationPreset] = {
        PluginPresetId.DEFAULT.value: BilaterationPreset(
            detector_config=get_default_detector_config(),
            bilateration_config=ProcessorConfig(),
        )
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._detector_instance: Optional[Detector] = None

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])
        self.shared_state.bilateration_config = ProcessorConfig.from_json(
            file["bilateration_config"][()]
        )
        self.shared_state.context = opser.deserialize(file["context"], DetectorContext)

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState(config=get_default_detector_config())
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            # Try to use the sensor ids from the last calibration
            if self.shared_state.context.sensor_ids:
                self.shared_state.sensor_ids = self.shared_state.context.sensor_ids

            for i in range(len(self.shared_state.sensor_ids)):
                if len(sensor_ids) > 0 and self.shared_state.sensor_ids[i] not in sensor_ids:
                    self.shared_state.sensor_ids[i] = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_processor_config(self, *, config: ProcessorConfig) -> None:
        self.shared_state.bilateration_config = config
        self.broadcast()

    @is_task
    def update_sensor_ids(self, *, sensor_ids: list[int]) -> None:
        self.shared_state.sensor_ids = sensor_ids
        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config.detector_config
        self.shared_state.bilateration_config = preset_config.bilateration_config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        _create_h5_string_dataset(
            file, "bilateration_config", self.shared_state.bilateration_config.to_json()
        )
        context_group = file.create_group("context")
        opser.serialize(self.shared_state.context, context_group)

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context
        self.shared_state.sensor_ids = list(next(iter(record.session_config.groups)).keys())

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_ids=self.shared_state.sensor_ids,
            detector_config=self.shared_state.config,
            context=self.shared_state.context,
        )

        if recorder:
            algo_group = recorder.require_algo_group("bilateration")
        else:
            algo_group = None

        self._processor_instance = Processor(
            session_config=self._detector_instance.session_config,
            processor_config=self.shared_state.bilateration_config,
            sensor_ids=self.shared_state.sensor_ids,
        )

        self._detector_instance.start(
            recorder,
            _algo_group=algo_group,
        )

        self.callback(
            SetupMessage(
                bilateration_config=self.shared_state.bilateration_config,
                num_curves=len(self._detector_instance.processor_specs),
                detector_config=self.shared_state.config,
            )
        )

    def end_session(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._detector_instance.stop()

    def get_next(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        detector_result = self._detector_instance.get_next()

        processor_result = self._processor_instance.process(result=detector_result)

        assert self.client is not None
        self.callback(backend.PlotMessage(result=(detector_result, processor_result)))

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._detector_instance = Detector(
                client=self.client,
                sensor_ids=self.shared_state.sensor_ids,
                detector_config=self.shared_state.config,
                context=None,
            )
            self._detector_instance.calibrate_detector()
        except Exception as exc:
            msg = "Failed to calibrate detector"
            raise HandledException(msg) from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._detector_instance.context
        self.broadcast()


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[tuple[Dict[int, DetectorResult], ProcessorResult]] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                message.bilateration_config,
                message.num_curves,
                message.detector_config,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            (detector_result, processor_result) = self._plot_job
            self.draw_plot_job(detector_result=detector_result, processor_result=processor_result)
        finally:
            self._plot_job = None

    def setup(
        self,
        bilateration_config: ProcessorConfig,
        num_curves: int,
        detector_config: DetectorConfig,
    ) -> None:
        self.plot_layout.clear()

        self.num_curves = num_curves
        self.detector_config = detector_config
        self.sensor_half_spacing_m = bilateration_config.sensor_spacing_m / 2

        win = self.plot_layout

        # Define sweep plot.
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        pen_sweep_0 = et.utils.pg_pen_cycler(0)
        pen_sweep_1 = et.utils.pg_pen_cycler(1)

        # Add sweep curves for the two sensors.
        feat_kw_1 = dict(pen=pen_sweep_0)
        feat_kw_2 = dict(pen=pen_sweep_1)
        self.sweep_curves_1 = [self.sweep_plot.plot(**feat_kw_1) for _ in range(self.num_curves)]
        self.sweep_curves_2 = [self.sweep_plot.plot(**feat_kw_2) for _ in range(self.num_curves)]

        pen_sweep_0 = et.utils.pg_pen_cycler(0, "--")
        pen_sweep_1 = et.utils.pg_pen_cycler(1, "--")

        # Add threshold curves for the two sensors.
        feat_kw_1 = dict(pen=pen_sweep_0)
        feat_kw_2 = dict(pen=pen_sweep_1)
        self.threshold_curves_1 = [
            self.sweep_plot.plot(**feat_kw_1) for _ in range(self.num_curves)
        ]
        self.threshold_curves_2 = [
            self.sweep_plot.plot(**feat_kw_2) for _ in range(self.num_curves)
        ]

        # Add legends.
        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweep_curves_1[0], "Sweep - Left sensor")
        sweep_plot_legend.addItem(self.threshold_curves_1[0], "Threshold - Left sensor")
        sweep_plot_legend.addItem(self.sweep_curves_2[0], "Sweep - Right sensor")
        sweep_plot_legend.addItem(self.threshold_curves_2[0], "Threshold - Right sensor")

        self.sweep_smooth_max = et.utils.SmoothMax()

        # Define obstacle plot.
        self.obstacle_location_plot = win.addPlot(row=1, col=0)
        self.obstacle_location_plot.setMenuEnabled(False)
        self.obstacle_location_plot.setAspectLocked()
        self.obstacle_location_plot.showGrid(x=True, y=True)
        self.obstacle_location_plot.addLegend()
        self.obstacle_location_plot.setLabel("left", "Y (m)")
        self.obstacle_location_plot.setLabel("bottom", "X (m)")
        self.obstacle_location_plot.addItem(pg.PlotDataItem())
        self.obstacle_location_plot.setXRange(
            -self.detector_config.end_m, self.detector_config.end_m
        )
        self.obstacle_location_plot.setYRange(0.0, self.detector_config.end_m)

        pen_sweep_0 = et.utils.pg_pen_cycler(0)
        brush_sweep_0 = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_sweep_0, symbolPen=None)
        feat_kw = dict(pen=None, **symbol_kw)
        self.obstacle_location_curve = self.obstacle_location_plot.plot(**feat_kw)
        feat_kw = dict(pen=pen_sweep_0)
        self.obstacle_location_half_curve = [
            self.obstacle_location_plot.plot(**feat_kw) for _ in range(Processor._MAX_NUM_OBJECTS)
        ]

    def draw_plot_job(
        self, *, detector_result: Dict[int, DetectorResult], processor_result: ProcessorResult
    ) -> None:
        # Plot sweep data from both distance detectors.
        max_val = 0.0
        for distance_detector_result, sweep_curves, threshold_curves in zip(
            detector_result.values(),
            [self.sweep_curves_1, self.sweep_curves_2],
            [self.threshold_curves_1, self.threshold_curves_2],
        ):
            for idx, distance_processor_result in enumerate(
                distance_detector_result.processor_results
            ):
                abs_sweep = distance_processor_result.extra_result.abs_sweep
                threshold = distance_processor_result.extra_result.used_threshold
                distances_m = distance_processor_result.extra_result.distances_m

                assert abs_sweep is not None
                assert threshold is not None
                assert distances_m is not None

                sweep_curves[idx].setData(distances_m, abs_sweep)
                threshold_curves[idx].setData(distances_m, threshold)

                if max_val < np.max(abs_sweep):
                    max_val = float(np.max(abs_sweep))

                if max_val < np.max(threshold):
                    max_val = float(np.max(threshold))

        if max_val != 0.0:
            self.sweep_plot.setYRange(0.0, self.sweep_smooth_max.update(max_val))

        # Plot result from bilateration processor.
        # Start with the points.
        points = processor_result.points
        xs = [point.x_coord for point in points]
        ys = [point.y_coord for point in points]
        self.obstacle_location_curve.setData(xs, ys)

        # Plot objects without counter part.
        objects_without_counterpart = processor_result.objects_without_counterpart
        num_points_on_circle = 100
        for i, o in enumerate(objects_without_counterpart):
            x = np.cos(np.linspace(0, np.pi, num_points_on_circle)) * o.distance
            y = np.sin(np.linspace(0, np.pi, num_points_on_circle)) * o.distance

            # Offset circle to center around the sensor that detected the object.
            if o.sensor_position == Processor._SENSOR_POSITION_LEFT:
                x -= self.sensor_half_spacing_m
            elif o.sensor_position == Processor._SENSOR_POSITION_RIGHT:
                x += self.sensor_half_spacing_m
            else:
                msg = "Invalid sensor position."
                raise ValueError(msg)

            self.obstacle_location_half_curve[i].setData(x, y)

        # Remove curves that does not have an object to visualize.
        for i in range(len(objects_without_counterpart), Processor._MAX_NUM_OBJECTS):
            self.obstacle_location_half_curve[i].setData([], [])


class ViewPlugin(A121ViewPluginBase):
    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.END_LESSER_THAN_START: "'Range end' point must be greater than 'Range "
        + "start'.",
        DetailedStatus.SENSOR_IDS_NOT_UNIQUE: "Select two different sensor IDs.",
        DetailedStatus.CONTEXT_MISSING: "Run detector calibration.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
    }

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_detector_button = QPushButton(icons.CALIBRATE(), "Calibrate detector")
        self.calibrate_detector_button.clicked.connect(self._on_calibrate_detector)

        self.defaults_button = QPushButton(icons.RESTORE(), "Reset settings and calibrations")
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_detector_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.defaults_button, 3, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 4, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.grid("Sensor selection", parent=self.scrolly_widget)
        self.two_sensor_id_editor = TwoSensorIdsEditor(name_label_texts=["Left", "Right"])
        sensor_selection_group.layout().addWidget(self.two_sensor_id_editor, 0, 0)
        scrolly_layout.addWidget(sensor_selection_group)

        self.bilateration_config_editor = AttrsConfigEditor(
            title="Bilateration parameters",
            factory_mapping=self._get_processor_pidget_mapping(),
            config_type=ProcessorConfig,
            parent=self.scrolly_widget,
        )
        self.bilateration_config_editor.sig_update.connect(self._on_processor_config_update)
        scrolly_layout.addWidget(self.bilateration_config_editor)

        self.config_editor = AttrsConfigEditor(
            title="Distance detector parameters",
            factory_mapping=self._get_pidget_mapping(),
            config_type=DetectorConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.two_sensor_id_editor.sig_update.connect(self._on_sensor_ids_update)
        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_processor_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "sensor_spacing_m": pidgets.FloatPidgetFactory(
                name_label_text="Sensor spacing:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "sensor_spacing_m"),
                suffix=" m",
                decimals=3,
            ),
            "max_anticipated_velocity_mps": pidgets.FloatPidgetFactory(
                name_label_text="Max anticipated velocity:",
                name_label_tooltip=get_attribute_docstring(
                    ProcessorConfig, "max_anticipated_velocity_mps"
                ),
                suffix=" m/s",
                decimals=1,
            ),
            "dead_reckoning_duration_s": pidgets.FloatPidgetFactory(
                name_label_text="Dead reckoning duration:",
                name_label_tooltip=get_attribute_docstring(
                    ProcessorConfig, "dead_reckoning_duration_s"
                ),
                suffix=" s",
                decimals=1,
            ),
            "sensitivity": pidgets.FloatSliderPidgetFactory(
                name_label_text="Sensitivity:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "sensitivity"),
                decimals=2,
                limits=(0.001, 1),
                show_limit_values=False,
                limit_texts=("Higher robustness", "More Responsive"),
            ),
        }

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetFactoryMapping:
        distance_pidget_mapping = dict(DistanceViewPlugin.get_pidget_mapping())

        # Bilateration requires an update rate, so replace the optional float with a mandatory
        distance_pidget_mapping["update_rate"] = pidgets.FloatPidgetFactory(
            name_label_text=distance_pidget_mapping["update_rate"].name_label_text,
            name_label_tooltip=distance_pidget_mapping["update_rate"].name_label_tooltip,
            decimals=1,
            limits=(1, None),
        )
        return distance_pidget_mapping

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.bilateration_config_editor.set_data(None)
            self.message_box.setText("")
        else:
            self.config_editor.set_data(state.config)
            self.bilateration_config_editor.set_data(state.bilateration_config)
            self.two_sensor_id_editor.set_data(state.sensor_ids)

            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )

            self.message_box.setText(self.TEXT_MSG_MAP[detector_status.detector_state])

            session_config = distance.detector_config_to_session_config(
                state.config, state.sensor_ids
            )

            validation_results = (
                state.config._collect_validation_results()
                + state.bilateration_config._collect_validation_results(session_config)
            )

            not_handled = self.config_editor.handle_validation_results(validation_results)
            not_handled = self.bilateration_config_editor.handle_validation_results(not_handled)
            not_handled = self.misc_error_view.handle_validation_results(not_handled)
            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.two_sensor_id_editor.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[
                self.config_editor,
                self.defaults_button,
                self.two_sensor_id_editor,
                self.bilateration_config_editor,
            ],
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

        state = app_model.backend_plugin_state

        if state is None:
            detector_ready = False
            state_valid = False
        else:
            detector_status = Detector.get_detector_status(
                state.config, state.context, state.sensor_ids
            )
            detector_ready = detector_status.ready_to_start

            state_valid = (
                self._config_valid(state)
                and self.config_editor.is_ready
                and self.bilateration_config_editor.is_ready
                and detector_status.detector_state is not DetailedStatus.SENSOR_IDS_NOT_UNIQUE
            )

        self.calibrate_detector_button.setEnabled(
            visual_policies.start_button_enabled(app_model, extra_condition=state_valid)
        )
        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=state_valid and detector_ready
            )
        )

    def _config_valid(self, state: SharedState) -> bool:
        session_config = distance.detector_config_to_session_config(state.config, state.sensor_ids)

        try:
            state.bilateration_config.validate(session_config)
            state.config.validate()
        except a121.ValidationResult:
            return False
        else:
            return True

    def _on_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _on_processor_config_update(self, config: ProcessorConfig) -> None:
        BackendPlugin.update_processor_config.rpc(self.app_model.put_task, config=config)

    def _on_sensor_ids_update(self, sensor_ids: list[int]) -> None:
        BackendPlugin.update_sensor_ids.rpc(self.app_model.put_task, sensor_ids=sensor_ids)

    def _on_calibrate_detector(self) -> None:
        BackendPlugin.calibrate_detector.rpc(self.app_model.put_task)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


BILATERATION_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="bilateration",
    title="Bilateration",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/bilateration.html",
    description="Use two sensors to estimate distance and angle.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import typing as t

from acconeer.exptool import a121

from .preview import PresentationType


_DISCLAIMER = """
This snippet is generated to be compatible with RSS A121 v1.0.0
If there is a version missmatch the snippet might need some modification
"""


def _sensor_config_parameters(config: a121.SensorConfig) -> str:
    return "\n".join(
        f"acc_config_{attr}_set(config, {value});"
        for attr, value in {
            "sweeps_per_frame": _uint(config.sweeps_per_frame),
            "sweep_rate": _rate(config.sweep_rate),
            "frame_rate": _rate(config.frame_rate),
            "inter_sweep_idle_state": _idle_state(config.inter_sweep_idle_state),
            "inter_frame_idle_state": _idle_state(config.inter_frame_idle_state),
            "continuous_sweep_mode": _bool(config.continuous_sweep_mode),
            "double_buffering": _bool(config.double_buffering),
        }.items()
    )


def _subsweep_parameters(config: a121.SubsweepConfig, *, index: t.Optional[int]) -> str:
    return "\n".join(
        [
            (
                f"acc_config_{attr}_set(config, {value});"
                if index is None
                else f"acc_config_subsweep_{attr}_set(config, {value}, {_uint(index)});"
            )
            for attr, value in {
                "start_point": config.start_point,
                "num_points": _uint(config.num_points),
                "step_length": _uint(config.step_length),
                "profile": _profile(config.profile),
                "hwaas": _uint(config.hwaas),
                "receiver_gain": _uint(config.receiver_gain),
                "enable_tx": _bool(config.enable_tx),
                "enable_loopback": _bool(config.enable_loopback),
                "phase_enhancement": _bool(config.phase_enhancement),
                "iq_imbalance_compensation": _bool(config.iq_imbalance_compensation),
                "prf": _prf(config.prf),
            }.items()
        ]
    )


def _idle_state(idle_state: a121.IdleState) -> str:
    return f"ACC_CONFIG_IDLE_STATE_{idle_state.name}".upper()


def _prf(prf: a121.PRF) -> str:
    return f"ACC_CONFIG_{prf.name}".upper()


def _profile(profile: a121.Profile) -> str:
    return f"ACC_CONFIG_{profile.name}"


def _uint(number: int) -> str:
    return f"{number}U"


def _rate(number: t.Optional[float]) -> str:
    return "0.0f" if number is None else f"{number:.3f}f"


def _bool(flag: bool) -> str:
    return str(flag).lower()


def _indent(body: str, indent: str = "\t") -> str:
    return indent + body.strip().replace("\n", "\n" + indent)


def _frame_it(function_body: str) -> str:
    return (
        "static void set_config(acc_config_t *config)\n"
        + "{\n"
        + _indent(_DISCLAIMER, indent="\t// ")
        + "\n"
        + "\n"
        + _indent(function_body)
        + "\n"
        + "}\n"
    )


def set_config_presenter(instance: t.Any, t: PresentationType) -> t.Optional[str]:
    if t is not PresentationType.C_SET_CONFIG:
        return None

    if isinstance(instance, a121.SessionConfig):
        try:
            instance = instance.sensor_config
        except RuntimeError:
            return None

    if isinstance(instance, a121.SensorConfig):
        if instance.num_subsweeps == 1:
            return _frame_it(
                _subsweep_parameters(instance.subsweep, index=None)
                + "\n"
                + "\n"
                + _sensor_config_parameters(instance)
            )
        else:
            return _frame_it(
                "\n\n".join(
                    [
                        _sensor_config_parameters(instance),
                        f"acc_config_num_subsweeps_set(config, {_uint(instance.num_subsweeps)});",
                    ]
                    + [
                        (f"// Subsweep {i}\n" + _subsweep_parameters(ssc, index=i))
                        for i, ssc in enumerate(instance.subsweeps)
                    ]
                )
            )

    return None



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from ._configs import get_infant_config, get_sitting_config
from ._processor import (
    AppState,
    BreathingProcessorConfig,
    Processor,
    ProcessorConfig,
    ProcessorResult,
)
from ._ref_app import RefApp, RefAppConfig, RefAppResult



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.app.new.ui.components import PidgetGroupFactoryMapping, pidgets
from acconeer.exptool.app.new.ui.components.pidgets.hooks import (
    disable_if,
    parameter_is,
)

from ._detector import DetectorConfig


def get_pidget_mapping() -> PidgetGroupFactoryMapping:
    service_parameters = {
        "start_m": pidgets.FloatPidgetFactory(
            name_label_text="Range start:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "start_m"),
            suffix=" m",
            decimals=3,
        ),
        "end_m": pidgets.FloatPidgetFactory(
            name_label_text="Range end:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "end_m"),
            suffix=" m",
            decimals=3,
        ),
        "frame_rate": pidgets.FloatPidgetFactory(
            name_label_text="Frame rate:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "frame_rate"),
            suffix=" Hz",
            decimals=1,
            limits=(0.1, 100),
        ),
        "sweeps_per_frame": pidgets.IntPidgetFactory(
            name_label_text="Sweeps per frame:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "sweeps_per_frame"),
            limits=(1, 4095),
        ),
        "inter_frame_idle_state": pidgets.EnumPidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter frame idle state:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "inter_frame_idle_state"),
            label_mapping={
                a121.IdleState.DEEP_SLEEP: "Deep sleep",
                a121.IdleState.SLEEP: "Sleep",
                a121.IdleState.READY: "Ready",
            },
        ),
    }
    signal_parameters = {
        "signal_quality": pidgets.FloatSliderPidgetFactory(
            name_label_text="Signal quality:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "signal_quality"),
            decimals=1,
            limits=(-10.0, 60.0),
        ),
    }
    subsweep_settings = {
        "hwaas": pidgets.IntPidgetFactory(
            name_label_text="HWAAS:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "hwaas"),
            limits=(1, 511),
        ),
        "profile": pidgets.OptionalEnumPidgetFactory(
            name_label_text="Profile:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "profile"),
            checkbox_label_text="Override",
            enum_type=a121.Profile,
            label_mapping={
                a121.Profile.PROFILE_1: "1 (shortest)",
                a121.Profile.PROFILE_2: "2",
                a121.Profile.PROFILE_3: "3",
                a121.Profile.PROFILE_4: "4",
                a121.Profile.PROFILE_5: "5 (longest)",
            },
        ),
        "step_length": pidgets.OptionalIntPidgetFactory(
            name_label_text="Step length:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "step_length"),
            checkbox_label_text="Override",
            limits=(1, None),
            init_set_value=24,
        ),
    }
    intra_parameters = {
        "intra_detection_threshold": pidgets.FloatSliderPidgetFactory(
            name_label_text="Intra detection threshold:",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "intra_detection_threshold"
            ),
            decimals=2,
            limits=(0, 100),
        ),
        "intra_frame_time_const": pidgets.FloatSliderPidgetFactory(
            name_label_text="Intra time constant:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "intra_frame_time_const"),
            suffix=" s",
            decimals=2,
            limits=(0, 1),
        ),
        "intra_output_time_const": pidgets.FloatSliderPidgetFactory(
            name_label_text="Intra output time constant:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "intra_output_time_const"),
            suffix=" s",
            decimals=2,
            limits=(0.01, 20),
            log_scale=True,
        ),
    }
    inter_parameters = {
        "inter_detection_threshold": pidgets.FloatSliderPidgetFactory(
            name_label_text="Inter detection threshold:",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "inter_detection_threshold"
            ),
            decimals=2,
            limits=(0, 100),
        ),
        "inter_frame_fast_cutoff": pidgets.FloatSliderPidgetFactory(
            name_label_text="Inter fast cutoff freq.:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "inter_frame_fast_cutoff"),
            suffix=" Hz",
            decimals=2,
            limits=(1, 50),
            log_scale=True,
        ),
        "inter_frame_slow_cutoff": pidgets.FloatSliderPidgetFactory(
            name_label_text="Inter slow cutoff freq.:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "inter_frame_slow_cutoff"),
            suffix=" Hz",
            decimals=2,
            limits=(0.01, 1),
            log_scale=True,
        ),
        "inter_frame_deviation_time_const": pidgets.FloatSliderPidgetFactory(
            name_label_text="Inter time constant:",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "inter_frame_deviation_time_const"
            ),
            suffix=" s",
            decimals=2,
            limits=(0.01, 20),
            log_scale=True,
        ),
        "inter_output_time_const": pidgets.FloatSliderPidgetFactory(
            name_label_text="Inter output time constant:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "inter_output_time_const"),
            suffix=" s",
            decimals=2,
            limits=(0.01, 20),
            log_scale=True,
        ),
        "inter_frame_presence_timeout": pidgets.OptionalIntPidgetFactory(
            name_label_text="Presence timeout:",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "inter_frame_presence_timeout"
            ),
            checkbox_label_text="Enable",
            suffix=" s",
            limits=(1, 30),
            init_set_value=5,
        ),
    }
    return {
        pidgets.FlatPidgetGroup(): service_parameters,
        pidgets.FlatPidgetGroup(): {
            "automatic_subsweeps": pidgets.CheckboxPidgetFactory(
                name_label_text="Enable automatic subsweep config (recommended)",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "automatic_subsweeps"),
            ),
        },
        pidgets.FlatPidgetGroup(
            hooks=disable_if(parameter_is("automatic_subsweeps", False)),
        ): signal_parameters,
        pidgets.FlatPidgetGroup(
            hooks=disable_if(parameter_is("automatic_subsweeps", True)),
        ): subsweep_settings,
        pidgets.FlatPidgetGroup(): {
            "intra_enable": pidgets.CheckboxPidgetFactory(
                name_label_text="Enable intra (fast) motion detection",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "intra_enable"),
            ),
        },
        pidgets.FlatPidgetGroup(
            hooks=disable_if(parameter_is("intra_enable", False)),
        ): intra_parameters,
        pidgets.FlatPidgetGroup(): {
            "inter_enable": pidgets.CheckboxPidgetFactory(
                name_label_text="Enable inter (slow) motion detection",
                name_label_tooltip=get_attribute_docstring(DetectorConfig, "inter_enable"),
            ),
        },
        pidgets.FlatPidgetGroup(
            hooks=disable_if(parameter_is("inter_enable", False)),
        ): inter_parameters,
    }



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from acconeer.exptool.flash._bootloader_tool import (
    XC120_BOOTLOADER_PID,
    XC120_EXPLORATION_SERVER_WINUSB_PID,
    BootLoaderXC120,
)
from acconeer.exptool.flash._mcumgruart import ACCONEER_XB122_MODULE_PID, McuMgrUartFlasher
from acconeer.exptool.flash._stm32uart import ACCONEER_XM_CP2105_MODULE_PID, Stm32UartFlasher


PRODUCT_PID_TO_FLASH_MAP = {
    XC120_BOOTLOADER_PID: BootLoaderXC120,
    XC120_EXPLORATION_SERVER_WINUSB_PID: BootLoaderXC120,
    ACCONEER_XM_CP2105_MODULE_PID: Stm32UartFlasher,
    ACCONEER_XB122_MODULE_PID: McuMgrUartFlasher,
}

PRODUCT_NAME_TO_FLASH_MAP = {
    "XC120": BootLoaderXC120,
    "XE125": Stm32UartFlasher,
    "XM125": Stm32UartFlasher,
    "XB122": McuMgrUartFlasher,
    "XM126": McuMgrUartFlasher,
}

EVK_TO_PRODUCT_MAP = {
    "XC120": "XC120",
    "XE125": "XM125",
    "XM125": "XM125",
    "XB122": "XM126",
    "XM126": "XM126",
}

EVKS_THAT_HAVE_FW_ON_DEVSITE = ["XC120", "XM125", "XM126"]



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from ._processor import Processor, ProcessorConfig, ProcessorLevelStatus, ProcessorResult
from ._ref_app import RefApp, RefAppConfig, RefAppContext, RefAppResult



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

from importlib.resources import as_file, files
from typing import Optional

from typing_extensions import Literal

from PySide6.QtCore import QEasingCurve, QEvent, QObject, QSize, Qt, QVariantAnimation, Signal
from PySide6.QtGui import QColor
from PySide6.QtWidgets import (
    QApplication,
    QGraphicsDropShadowEffect,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

import acconeer.exptool.app.resources as et_app_resources


class ImageButton(QToolButton):
    def __init__(
        self, text: str, image_path: str, size: QSize, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)
        self._layout = QVBoxLayout(self)
        self.setLayout(self._layout)
        self.setFixedSize(size)
        self.setIconSize(size)

        self._image_path = image_path
        self._animator = QVariantAnimation(self)
        self._animator.valueChanged.connect(self._update_zoom)

        label_shadow = QGraphicsDropShadowEffect(self)
        label_shadow.setColor(QColor(64, 64, 64))
        label_shadow.setOffset(2, 2)
        label_shadow.setBlurRadius(10)

        label = QLabel(text)
        label.setGraphicsEffect(label_shadow)

        self._layout.addStretch()
        self._layout.addWidget(label, 0, alignment=Qt.AlignmentFlag.AlignCenter)
        self.setStyleSheet(
            f"""
            QLabel {{
                font: bold;
                color: #000;
                font-size: 32px;
            }}
            QToolButton {{
                /* The border-image attribute is hack:
                 * https://forum.qt.io/topic/40151/solved-scaled-background-image-using-stylesheet
                 */
                border-image: url('{image_path}') 0 0 0 0 stretch stretch;
            }}
            """
        )
        self.installEventFilter(self)

    def _do_animation(self, start_value: int, end_value: int, duration: int = 500) -> None:
        self._animator.stop()
        self._animator.setDuration(duration)
        self._animator.setStartValue(start_value)
        self._animator.setEndValue(end_value)
        self._animator.setEasingCurve(QEasingCurve.Type.InOutSine)
        self._animator.start()

    def eventFilter(self, _: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Type.HoverEnter:
            self._do_animation(0, 100, duration=300)

        if event.type() == QEvent.Type.HoverLeave:
            self._do_animation(100, 0, duration=100)

        return False

    def _update_zoom(self, current_value: float) -> None:
        zoom_string = str(current_value)
        self.setStyleSheet(
            f"""
            QLabel {{
                font: bold;
                color: #000;
                font-size: 32px;
            }}
            QToolButton {{
                border-image: url('{self._image_path}') {zoom_string} stretch stretch;
            }}
            """
        )


class CentralWidget(QWidget):
    sig_new_app_clicked = Signal()
    sig_old_app_clicked = Signal()

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        layout = QHBoxLayout(self)

        with as_file(files(et_app_resources) / "new_gui.png") as path:
            new_app_button = ImageButton("A121", path.as_posix(), QSize(500, 300))
            new_app_button.clicked.connect(self.sig_new_app_clicked)
            layout.addWidget(new_app_button)

        with as_file(files(et_app_resources) / "old_gui.png") as path:
            old_app_button = ImageButton("A111", path.as_posix(), QSize(500, 300))
            old_app_button.clicked.connect(self.sig_old_app_clicked)
            layout.addWidget(old_app_button)

        self.setLayout(layout)


class ValueHolder:
    def __init__(self) -> None:
        self.selection: Optional[Literal["new", "old"]] = None


class Launcher(QMainWindow):
    def __init__(self, value_holder: ValueHolder) -> None:
        super().__init__()
        self._value_holder = value_holder

        self._central_widget = CentralWidget()
        self._central_widget.sig_new_app_clicked.connect(lambda: self.on_selection("new"))
        self._central_widget.sig_old_app_clicked.connect(lambda: self.on_selection("old"))

        self.setWindowTitle("Acconeer Exptool Launcher")
        self.setCentralWidget(self._central_widget)

    def on_selection(self, selection: Literal["new", "old"]) -> None:
        self._value_holder.selection = selection
        self.close()


def run_launcher() -> Optional[Literal["new", "old"]]:
    QApplication.setHighDpiScaleFactorRoundingPolicy(
        Qt.HighDpiScaleFactorRoundingPolicy.PassThrough,
    )

    app = QApplication([])

    vh = ValueHolder()
    launcher = Launcher(vh)
    launcher.show()

    app.exec()

    return vh.selection



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from typing import Iterator, Optional, Sequence

import attrs

from acconeer.exptool._core.entities import ClientInfo
from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities import (
    Metadata,
    Record,
    RecordException,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
    SessionRecord,
    StackedResults,
)


class InMemoryRecordException(Exception):
    pass


@attrs.frozen(kw_only=True)
class InMemorySessionRecord(SessionRecord):
    extended_metadata: list[dict[int, Metadata]]
    extended_stacked_results: list[dict[int, StackedResults]]
    num_frames: int
    session_config: SessionConfig
    _calibrations: Optional[dict[int, SensorCalibration]]
    _calibrations_provided: Optional[dict[int, bool]]

    @property
    def extended_results(self) -> Iterator[list[dict[int, Result]]]:
        for frame_no in range(self.num_frames):
            yield self._get_result_for_all_entries(frame_no)

    @property
    def sensor_id(self) -> int:
        return self.session_config.sensor_id

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        if self._calibrations is None:
            msg = "No calibration in record"
            raise InMemoryRecordException(msg)
        return self._calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        if self._calibrations_provided is None:
            msg = "No calibration in record"
            raise InMemoryRecordException(msg)
        return self._calibrations_provided

    def _get_result_for_all_entries(self, frame_no: int) -> list[dict[int, Result]]:
        def stacked_results_to_result(stacked_results: StackedResults) -> Result:
            return stacked_results[frame_no]

        return utils.map_over_extended_structure(
            stacked_results_to_result, self.extended_stacked_results
        )

    @classmethod
    def from_session_record(cls, session_record: SessionRecord) -> InMemorySessionRecord:
        try:
            calibrations = session_record.calibrations
            calibrations_provided = session_record.calibrations_provided
        except RecordException:
            calibrations = None
            calibrations_provided = None

        return cls(
            extended_metadata=session_record.extended_metadata,
            extended_stacked_results=session_record.extended_stacked_results,
            num_frames=session_record.num_frames,
            session_config=session_record.session_config,
            calibrations=calibrations,
            calibrations_provided=calibrations_provided,
        )


@attrs.frozen(kw_only=True)
class InMemoryRecord(Record):
    client_info: ClientInfo
    lib_version: str
    server_info: ServerInfo
    timestamp: str
    uuid: str
    _sessions: Sequence[InMemorySessionRecord]

    def session(self, session_index: int) -> InMemorySessionRecord:
        return self._sessions[session_index]

    @property
    def num_sessions(self) -> int:
        return len(self._sessions)

    @classmethod
    def from_record(cls, record: Record) -> InMemoryRecord:
        return cls(
            client_info=record.client_info,
            lib_version=record.lib_version,
            server_info=record.server_info,
            timestamp=record.timestamp,
            uuid=record.uuid,
            sessions=tuple(
                InMemorySessionRecord.from_session_record(record.session(i))
                for i in range(record.num_sessions)
            ),
        )



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from ._configs import (
    get_ground_config,
    get_pole_config,
)
from ._processors import (
    ObstructionProcessor,
    ObstructionProcessorConfig,
    Processor,
    ProcessorConfig,
    ProcessorExtraResult,
    ProcessorResult,
)
from ._ref_app import RefApp, RefAppConfig, RefAppResult



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np
from scipy import signal

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.IQServiceConfig()
    config.range_interval = [0.4, 0.8]
    config.update_rate = 60
    config.gain = 0.6
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    n_dft = et.configbase.FloatParameter(
        label="Estimation window",
        unit="s",
        default_value=15,
        limits=(2, 20),
        updateable=False,
        order=0,
    )

    t_freq_est = et.configbase.FloatParameter(
        label="Time between estimation",
        unit="s",
        default_value=0.2,
        limits=(0.1, 10),
        updateable=False,
        order=10,
    )

    D = et.configbase.IntParameter(
        label="Distance downsampling",
        default_value=124,
        limits=(0, 248),
        updateable=False,
        order=20,
    )

    f_high = et.configbase.FloatParameter(
        label="Bandpass high freq",
        unit="Hz",
        default_value=0.8,
        limits=(0, 10),
        updateable=False,
        order=30,
    )

    f_low = et.configbase.FloatParameter(
        label="Bandpass low freq",
        unit="Hz",
        default_value=0.2,
        limits=(0, 10),
        updateable=False,
        order=40,
    )

    lambda_p = et.configbase.FloatParameter(
        label="Threshold: Peak to noise ratio",
        default_value=40,
        limits=(1, 1000),
        updateable=False,
        order=50,
    )

    lambda_05 = et.configbase.FloatParameter(
        label="Threshold: Peak to half harmonic ratio",
        default_value=1,
        limits=(0, 10),
        updateable=False,
        order=60,
    )


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.config = sensor_config

        # Settings
        # Data length for frequency estimation [s] | 20
        n_dft = processing_config.n_dft
        # Time between frequency estimations [s] | 2
        t_freq_est = processing_config.t_freq_est
        # Time constant low-pass filter on IQ-data [s] | 0.04
        tau_iq = 0.04
        # Time constant low-pass filter on IQ-data [s] | 150
        self.f_s = self.config.update_rate
        # Spatial or Range down sampling factor | 124
        self.D = processing_config.D
        # Lowest frequency of interest [Hz] | 0.1
        self.f_low = processing_config.f_low
        # Highest frequency of interest [Hz] | 1
        self.f_high = processing_config.f_high
        # Time down sampling for DFT | 40 f_s/M ~ 10 Hz
        self.M = int(self.f_s / 10)
        # Threshold: spectral peak to noise ratio [1] | 50
        self.lambda_p = processing_config.lambda_p
        # Threshold: ratio fundamental and half harmonic
        self.lambda_05 = processing_config.lambda_05
        # Interpolation between DFT points
        self.interpolate = True

        self.delta_f = 1 / n_dft
        self.dft_f_vec = np.arange(self.f_low, self.f_high, self.delta_f)
        self.dft_points = np.size(self.dft_f_vec)

        # Butterworth bandpass filter
        f_n = self.f_s / 2
        v_low = self.f_low / f_n
        v_high = self.f_high / f_n
        self.b, self.a = signal.butter(4, [v_low, v_high], btype="bandpass")

        # Exponential lowpass filter
        self.alpha_iq = np.exp(-2 / (self.f_s * tau_iq))
        self.alpha_phi = np.exp(-2 * self.f_low / self.f_s)

        # Parameter init
        self.sweeps_in_block = int(np.ceil(n_dft * self.f_s))
        self.new_sweeps_per_results = int(np.ceil(t_freq_est * self.f_s))
        self.phi_vec = np.zeros((self.sweeps_in_block, 1))
        self.f_est_vec = np.zeros(1)
        self.f_dft_est_vec = np.zeros(1)
        self.snr_vec = 0

        self.sweep_index = 0

    def process(self, data, data_info):
        sweep = data

        if self.sweep_index == 0:
            delay_points = int(np.ceil(np.size(sweep) / self.D))
            self.data_s_d_mat = np.zeros((self.sweeps_in_block, delay_points), dtype="complex")
            self.data_s_d_mat[self.sweep_index, :] = self.downsample(sweep, self.D)

            out_data = None
        elif self.sweep_index < self.sweeps_in_block:
            self.data_s_d_mat[self.sweep_index, :] = self.iq_lp_filter_time(
                self.data_s_d_mat[self.sweep_index - 1, :], self.downsample(sweep, self.D)
            )

            temp_phi = self.unwrap_phase(
                self.phi_vec[self.sweep_index - 1],
                self.data_s_d_mat[self.sweep_index, :],
                self.data_s_d_mat[self.sweep_index - 1, :],
            )

            self.phi_vec[self.sweep_index] = self.unwrap_phase(
                self.phi_vec[self.sweep_index - 1],
                self.data_s_d_mat[self.sweep_index, :],
                self.data_s_d_mat[self.sweep_index - 1, :],
            )

            phi_filt = signal.lfilter(self.b, self.a, self.phi_vec, axis=0)

            out_data = {
                "phi_raw": self.phi_vec,
                "phi_filt": phi_filt,
                "power_spectrum": np.zeros(self.dft_points),
                "x_dft": np.linspace(self.f_low, self.f_high, self.dft_points),
                "f_dft_est_hist": self.f_dft_est_vec,
                "f_est_hist": self.f_est_vec,
                "f_dft_est": 0,
                "f_est": 0,
                "f_low": self.f_low,
                "f_high": self.f_high,
                "snr": 0,
                "lambda_p": self.lambda_p,
                "lambda_05": self.lambda_05,
                "dist_range": self.config.range_interval,
                "init_progress": round(100 * self.sweep_index / self.sweeps_in_block),
            }
        else:
            # Lowpass filter IQ data downsampled in distance points
            self.data_s_d_mat = np.roll(self.data_s_d_mat, -1, axis=0)
            self.data_s_d_mat[-1, :] = self.iq_lp_filter_time(
                self.data_s_d_mat[-2, :], self.downsample(sweep, self.D)
            )

            # Phase unwrapping of IQ data
            temp_phi = self.unwrap_phase(
                self.phi_vec[-1], self.data_s_d_mat[-1, :], self.data_s_d_mat[-2, :]
            )
            self.phi_vec = np.roll(self.phi_vec, -1, axis=0)
            self.phi_vec[-1] = temp_phi

            if np.mod(self.sweep_index, self.new_sweeps_per_results - 1) == 0:
                # Bandpass filter unwrapped data
                phi_filt_vec = signal.lfilter(self.b, self.a, self.phi_vec, axis=0)
                P, dft_est, _ = self.dft(self.downsample(phi_filt_vec, self.M))
                f_breath_est, _, snr, _ = self.breath_freq_est(P)

                self.f_est_vec = np.append(self.f_est_vec, f_breath_est)
                self.f_dft_est_vec = np.append(self.f_dft_est_vec, dft_est)
                self.snr_vec = np.append(self.snr_vec, snr)

                out_data = {
                    "phi_raw": self.phi_vec,
                    "phi_filt": phi_filt_vec,
                    "power_spectrum": P,
                    "x_dft": np.linspace(self.f_low, self.f_high, self.dft_points),
                    "f_dft_est_hist": self.f_dft_est_vec,
                    "f_est_hist": self.f_est_vec,
                    "f_dft_est": dft_est,
                    "f_est": f_breath_est,
                    "f_low": self.f_low,
                    "f_high": self.f_high,
                    "snr": snr,
                    "lambda_p": self.lambda_p,
                    "lambda_05": self.lambda_05,
                    "dist_range": self.config.range_interval,
                    "init_progress": None,
                }
            else:
                out_data = None

        self.sweep_index += 1
        return out_data

    def downsample(self, data, n):
        return data[::n]

    def iq_lp_filter_time(self, state, new_data):
        return self.alpha_iq * state + (1 - self.alpha_iq) * new_data

    def unwrap_phase(self, phase_lp, data_1, data_2):
        return phase_lp * self.alpha_phi + np.angle(np.mean(data_2 * np.conjugate(data_1)))

    def dft(self, data):
        data = np.squeeze(data)
        n_vec = np.arange(data.size) * self.M
        dft = np.exp((2j * np.pi / self.f_s) * np.outer(self.dft_f_vec, n_vec))
        P = np.square(np.abs(np.matmul(dft, data)))
        idx_f = np.argmax(P)
        dft_est = self.dft_f_vec[idx_f]
        return P, dft_est, P[idx_f]

    def noise_est(self, P):
        return np.mean(np.sort(P)[: (self.dft_points // 2) - 1])

    def half_peak_frequency(self, P, f_est):
        idx_half = int(f_est / (2 * self.delta_f))
        if idx_half < self.f_low:
            return 0
        else:
            return (1 / self.delta_f) * (
                (self.dft_f_vec[idx_half + 1] - f_est / 2) * P[idx_half]
                + (f_est / 2 - self.dft_f_vec[idx_half]) * P[idx_half + 1]
            )

    def breath_freq_est(self, P):
        f_idx = np.argmax(P)
        P_peak = P[f_idx]

        if self.interpolate:
            f_est, P_peak = self.freq_quad_interpolation(P)
        else:
            f_est = self.dft_f_vec[f_idx]

        P_half = self.half_peak_frequency(P, f_est)

        if P_peak < self.lambda_05 * P_half:
            f_est = f_est / 2
            P_peak = P_half

        if self.f_low < f_est < self.f_high and P_peak > self.lambda_p * self.noise_est(P):
            f_est_valid = True
        else:
            f_est_valid = False
            f_est = 0

        snr = P_peak / self.noise_est(P)
        return f_est, P_peak, snr, f_est_valid

    def freq_quad_interpolation(self, P):
        f_idx = np.argmax(P)

        if 0 < f_idx < (P.size - 1) and P.size > 3:
            f_est = self.dft_f_vec[f_idx] + self.delta_f / 2 * (
                (np.log(P[f_idx + 1]) - np.log(P[f_idx - 1]))
                / (2 * np.log(P[f_idx]) - np.log(P[f_idx + 1]) - np.log(P[f_idx - 1]))
            )
            P_peak = P[f_idx] + np.exp(
                (1 / 8)
                * np.square(np.log(P[f_idx + 1]) - np.log(P[f_idx - 1]))
                / (2 * np.log(P[f_idx]) - np.log(P[f_idx + 1]) - np.log(P[f_idx - 1]))
            )

            if not (self.f_low < f_est < self.f_high):
                f_est = 0
        else:
            f_est = 0
            P_peak = 0

        return f_est, P_peak



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import typing as t
from enum import Enum


_EnumT = t.TypeVar("_EnumT", bound=Enum)


def find_by_value(enum: type[_EnumT], value: object) -> t.Optional[_EnumT]:
    """Helper for writing readable Enum._missing_ methods"""
    for member in enum:
        if member.value == value:
            return member
    return None


def find_by_first_element_in_value(enum: type[_EnumT], value: object) -> t.Optional[_EnumT]:
    """Helper for writing readable Enum._missing_ methods"""
    for member in enum:
        if member.value[0] == value:
            return member
    return None


def find_by_name(enum: type[_EnumT], name: object) -> t.Optional[_EnumT]:
    """Helper for writing readable Enum._missing_ methods"""
    for member in enum:
        if member.name == name:
            return member
    return None


def find_by_lowercase_name(enum: type[_EnumT], lowercase_name: object) -> t.Optional[_EnumT]:
    """Helper for writing readable Enum._missing_ methods"""
    for member in enum:
        if member.name.lower() == lowercase_name:
            return member
    return None



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from .backend_plugin import (
    ExtendedProcessorBackendPluginBase,
    GenericProcessorBackendPluginBase,
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
    SetupMessage,
)
from .plugin import ProcessorPluginSpec
from .view_plugin import ProcessorViewPluginBase



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

import enum
from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import (
    attrs_ndarray_isclose,
    attrs_optional_ndarray_isclose,
)
from acconeer.exptool._core.int_16_complex import complex_array_to_int16_complex
from acconeer.exptool.a121.algo import (
    PERCEIVED_WAVELENGTH,
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ProcessorBase,
    double_buffering_frame_filter,
)
from acconeer.exptool.utils import is_power_of_2


RANGE_SUBSWEEP = 0
LOOPBACK_SUBSWEEP = 1


class ReportedDisplacement(AlgoParamEnum):
    """Selects how displacement is reported."""

    AMPLITUDE = enum.auto()
    """Report displacement as amplitude."""

    PEAK2PEAK = enum.auto()
    """Report displacement as peak to peak."""


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    time_series_length: int = attrs.field(default=1024)
    """Length of time series.
    This value will be overridden by the number of sweeps per frame if continuous sweep mode is not enabled."""

    lp_coeff: float = attrs.field(default=0.95)
    """Specify filter coefficient of the exponential filter for the FFT.
    A higher value means more filtering, i.e., slower response to changes."""

    threshold_margin: float = attrs.field(default=10.0)
    """Specify threshold margin (micrometer)."""

    amplitude_threshold: float = attrs.field(default=100.0)
    """Specify minimum amplitude for calculating vibration."""

    reported_displacement_mode: ReportedDisplacement = attrs.field(
        default=ReportedDisplacement.AMPLITUDE,
        converter=ReportedDisplacement,
    )
    """Selects whether to report the amplitude or peak to peak of the estimated frequency."""

    low_frequency_enhancement: bool = attrs.field(default=False)
    """Adds a loopback subsweep for phase correction to enhance low frequency detection."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if not is_power_of_2(self.time_series_length):
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "time_series_length",
                    "Should be power of 2 for efficient usage of fast fourier transform",
                )
            )

        if self.time_series_length < config.sensor_config.sweeps_per_frame:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "time_series_length",
                    "Should be greater or equal to sweeps_per_frame",
                )
            )

        if config.sensor_config.sweep_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweep_rate",
                    "Must be set",
                )
            )

        if config.sensor_config.subsweeps[0].num_points != 1:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config.subsweeps[0],
                    "num_points",
                    "Num points must be set to 1 for the first subsweep",
                )
            )

        if config.sensor_config.num_subsweeps > 2:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "num_subsweeps",
                    "Num subsweeps must not be more than 2",
                )
            )

        if config.sensor_config.num_subsweeps == 2:
            if config.sensor_config.subsweeps[1].num_points != 1:
                validation_results.append(
                    a121.ValidationError(
                        config.sensor_config.subsweeps[1],
                        "num_points",
                        "Num points must be set to 1 for the second subsweep",
                    )
                )
            if not self.low_frequency_enhancement:
                validation_results.append(
                    a121.ValidationError(
                        self.low_frequency_enhancement,
                        "low_frequency_enhancement",
                        "Low frequency enhancement must be enabled when there are two subsweeps",
                    )
                )

        if (
            config.sensor_config.continuous_sweep_mode
            and not config.sensor_config.double_buffering
        ):
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "continuous_sweep_mode",
                    "Continuous sweep mode requires double buffering to be enabled",
                )
            )

        if self.low_frequency_enhancement:
            if config.sensor_config.num_subsweeps != 2:
                a121.ValidationError(
                    self.low_frequency_enhancement,
                    "low_frequency_enhancement",
                    "Low frequency enhancement requires a loopback subsweep",
                )
            if config.sensor_config.subsweeps[RANGE_SUBSWEEP].enable_loopback:
                a121.ValidationError(
                    self.low_frequency_enhancement,
                    "low_frequency_enhancement",
                    "The first subsweep must not have loopback enabled",
                )
            if not config.sensor_config.subsweeps[LOOPBACK_SUBSWEEP].enable_loopback:
                a121.ValidationError(
                    self.low_frequency_enhancement,
                    "low_frequency_enhancement",
                    "Low frequency enhancement requires that loopback is enabled for the second subsweep",
                )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext: ...


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    amplitude_threshold: float
    """Amplitude threshold."""

    zm_time_series: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    """Time series being analyzed."""

    lp_displacements_threshold: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    """CFAR threshold used for detecting significant frequencies in the FFT."""


@attrs.frozen(kw_only=True)
class ProcessorResult:
    max_sweep_amplitude: float
    """Max amplitude in sweep.

    Used to determine whether an object is in front of the sensor.
    """

    lp_displacements: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    """Array of estimated displacement (μm) per frequency."""

    lp_displacements_freqs: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    """Array of frequencies where displacement is estimated (Hz)."""

    max_displacement: Optional[float] = attrs.field(default=None)
    """Largest detected displacement (μm)."""

    max_displacement_freq: Optional[float] = attrs.field(default=None)
    """Frequency of largest detected displacement (Hz)."""

    time_series_std: Optional[float] = attrs.field(default=None)
    """Time series standard deviation."""

    extra_result: ProcessorExtraResult
    """Extra result, used for plotting only."""


class Processor(ProcessorBase[ProcessorResult]):
    _WINDOW_BASE_LENGTH = 10
    _HALF_GUARD_BASE_LENGTH = 5
    _CFAR_MARGIN = _WINDOW_BASE_LENGTH + _HALF_GUARD_BASE_LENGTH

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        # Check sensor config and processor config
        assert sensor_config.sweep_rate is not None
        processor_config.validate(sensor_config)

        # Constants
        self.continuous_data_acquisition = (
            sensor_config.double_buffering & sensor_config.continuous_sweep_mode
        )
        self.lp_coeffs = processor_config.lp_coeff
        self.sensitivity = processor_config.threshold_margin
        self.time_series_length = processor_config.time_series_length
        self.amplitude_threshold = processor_config.amplitude_threshold
        self.spf = sensor_config.sweeps_per_frame
        self.reported_displacement_mode = processor_config.reported_displacement_mode
        self.low_frequency_enhancement = processor_config.low_frequency_enhancement
        self.sensor_config = sensor_config

        if self.continuous_data_acquisition:
            self.psd_to_radians_conversion_factor = 2.0 / float(self.time_series_length)
        else:
            self.psd_to_radians_conversion_factor = 2.0 / float(self.spf)

        self.radians_to_displacement = PERCEIVED_WAVELENGTH * 10**6 / (2 * np.pi)

        self.freq = np.fft.rfftfreq(
            processor_config.time_series_length,
            1 / sensor_config.sweep_rate,
        )[1:]

        # Variables
        self.time_series = np.zeros(shape=processor_config.time_series_length)
        self.lp_displacements = np.zeros_like(self.freq)

        self.has_init = False

    def process(self, result: a121.Result) -> ProcessorResult:
        # Extract loopback frame
        if not self.low_frequency_enhancement:
            frame = result.subframes[RANGE_SUBSWEEP]
        else:
            measured_frame = result.subframes[RANGE_SUBSWEEP]
            loopback_frame = result.subframes[LOOPBACK_SUBSWEEP]
            frame = measured_frame * np.exp(-1j * np.angle(loopback_frame))

        # Determine if an object is in front of the sensor
        max_sweep_amplitude = float(np.max(np.abs(frame)))

        if max_sweep_amplitude < self.amplitude_threshold:
            self.has_init = False
            # No object found -> Return
            return ProcessorResult(
                max_sweep_amplitude=max_sweep_amplitude,
                lp_displacements_freqs=self.freq,
                extra_result=ProcessorExtraResult(
                    amplitude_threshold=self.amplitude_threshold,
                ),
            )

        # Handle frame based on whether or not continuous sweep mode is used
        if self.continuous_data_acquisition:
            filter_output = double_buffering_frame_filter(complex_array_to_int16_complex(frame))
            if filter_output is not None:
                frame = filter_output
            self.time_series = np.roll(self.time_series, -self.spf)
            self.time_series[-self.spf :] = np.angle(frame.squeeze(axis=1))
            self.time_series = np.unwrap(self.time_series)
        else:
            self.time_series = np.unwrap(np.angle(frame.squeeze(axis=1)))

        # Calculate zero mean time series
        zm_time_series = self.time_series - np.mean(self.time_series)

        # Estimate displacement per frequency
        z_abs = np.abs(
            np.fft.rfft(
                zm_time_series,
                n=self.time_series_length,
            )
        )[1:]

        if self.reported_displacement_mode is ReportedDisplacement.AMPLITUDE:
            displacements = (
                z_abs * self.psd_to_radians_conversion_factor * self.radians_to_displacement
            )
        elif self.reported_displacement_mode is ReportedDisplacement.PEAK2PEAK:
            displacements = (
                z_abs * self.psd_to_radians_conversion_factor * self.radians_to_displacement * 2.0
            )
        else:
            msg = "Invalid reported displacement"
            raise RuntimeError(msg)

        if not self.has_init:
            self.lp_displacements = displacements
            self.has_init = True
        else:
            self.lp_displacements = self.lp_displacements * self.lp_coeffs + displacements * (
                1 - self.lp_coeffs
            )

        # Convert time series to um and calculate std
        zm_time_series_um = zm_time_series * self.radians_to_displacement
        time_series_rms = np.sqrt(np.mean(zm_time_series_um**2))

        # Identify peaks in spectrum
        lp_displacements_threshold = self._calculate_cfar_threshold(
            self.lp_displacements,
            self.sensitivity,
            self._WINDOW_BASE_LENGTH,
            self._HALF_GUARD_BASE_LENGTH,
        )

        lp_displacements_threshold = self._extend_cfar_threshold(lp_displacements_threshold)

        # Compare displacements to threshold and exclude first point as it does not form a peak
        idx_over_threshold = (
            np.where(lp_displacements_threshold[1:] < self.lp_displacements[1:])[0] + 1
        )

        if len(idx_over_threshold) != 0:
            displacements_over_threshold = self.lp_displacements[idx_over_threshold]
            max_displacement = np.max(displacements_over_threshold)
            max_displacement_freq = self.freq[
                idx_over_threshold[np.argmax(displacements_over_threshold)]
            ]
        else:
            max_displacement = None
            max_displacement_freq = None

        return ProcessorResult(
            time_series_std=time_series_rms,
            lp_displacements_freqs=self.freq,
            lp_displacements=self.lp_displacements,
            max_sweep_amplitude=max_sweep_amplitude,
            max_displacement=max_displacement,
            max_displacement_freq=max_displacement_freq,
            extra_result=ProcessorExtraResult(
                zm_time_series=zm_time_series_um,
                amplitude_threshold=self.amplitude_threshold,
                lp_displacements_threshold=lp_displacements_threshold,
            ),
        )

    @classmethod
    def _extend_cfar_threshold(cls, threshold: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Extends CFAR threshold using extrapolation

        The head of the threshold is extended using linear extrapolation, based on the first points
        of the original threshold.

        The tail of the threshold is extended using the average of the last threshold values of the
        original threshold.
        """
        head_slope_multiplier = 2.0
        head_slope_calculation_width = 3
        tail_mean_calculation_width = 10

        # Extend head
        base_offset = threshold[cls._CFAR_MARGIN]
        mean_slope = (
            np.mean(
                np.diff(
                    threshold[cls._CFAR_MARGIN : cls._CFAR_MARGIN + head_slope_calculation_width]
                )
            )
            * head_slope_multiplier
        )
        threshold[: cls._CFAR_MARGIN] = base_offset + mean_slope * np.arange(-cls._CFAR_MARGIN, 0)

        # Extend tail
        threshold[-cls._CFAR_MARGIN :] = np.mean(
            threshold[-cls._CFAR_MARGIN - tail_mean_calculation_width : -cls._CFAR_MARGIN]
        )

        return threshold

    @staticmethod
    def _calculate_cfar_threshold(
        psd: npt.NDArray[np.float64],
        sensitivity: float,
        window_length: int,
        half_guard_length: int,
    ) -> npt.NDArray[np.float64]:
        threshold = np.full(psd.shape, np.nan)
        margin = window_length + half_guard_length
        length_after_filtering = psd.shape[0] - 2 * margin

        filt_psd = np.convolve(psd, np.ones(window_length), "valid") / window_length
        threshold[margin:-margin] = (
            filt_psd[:length_after_filtering] + filt_psd[-length_after_filtering:]
        ) / 2 + sensitivity
        return threshold



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components.a121 import RangeHelpView

from ._example_app import ExampleApp, ExampleAppConfig, ExampleAppResult, _load_algo_data


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: ExampleAppConfig = attrs.field(factory=ExampleAppConfig)


class PluginPresetId(Enum):
    DEFAULT = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    example_app_config: ExampleAppConfig
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], ExampleAppConfig]] = {
        PluginPresetId.DEFAULT.value: lambda: ExampleAppConfig()
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._exempel_app_instance: Optional[ExampleApp] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = ExampleAppConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: ExampleAppConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.sensor_id = record.sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._example_app_instance = ExampleApp(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            example_app_config=self.shared_state.config,
        )
        self._example_app_instance.start(recorder)
        self.callback(SetupMessage(example_app_config=self.shared_state.config))

    def end_session(self) -> None:
        if self._example_app_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._example_app_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._example_app_instance is None:
            raise RuntimeError
        result = self._example_app_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class PlotPlugin(PgPlotPlugin):
    _VELOCITY_Y_SCALE_MARGIN_M = 0.25

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[ExampleAppResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            if not isinstance(message.result, ExampleAppResult):
                msg = f"Unexpected result type: {type(message.result)}"
                raise RuntimeError(msg)

            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(example_app_config=message.example_app_config)
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        example_app_config: ExampleAppConfig,
    ) -> None:
        self.plot_layout.clear()

        self.slow_zone = example_app_config.slow_zone
        self.history_length_s = 10
        if example_app_config.frame_rate is None:
            estimated_frame_rate = (
                example_app_config.sweep_rate / example_app_config.sweeps_per_frame
            )
        else:
            estimated_frame_rate = example_app_config.frame_rate

        self.history_length_n = int(np.around(self.history_length_s * estimated_frame_rate))

        c0_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        # Velocity plot

        self.velocity_history_plot = self._create_plot(self.plot_layout, row=0, col=0)
        self.velocity_history_plot.setTitle("Estimated velocity")
        self.velocity_history_plot.setLabel(axis="left", text="Velocity", units="m/s")
        self.velocity_history_plot.setLabel(axis="bottom", text="Time", units="s")
        self.velocity_history_plot.addLegend(labelTextSize="10pt")
        self.velocity_smooth_limits = et.utils.SmoothLimits()

        self.velocity_curve = self.velocity_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0), name="Estimated velocity"
        )

        self.psd_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:13pt;">'
            "{}</span></div>"
        )

        self.distance_text_item = pg.TextItem(
            html=self.psd_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.velocity_history_plot.addItem(self.distance_text_item)

        self.velocity_history = np.zeros(self.history_length_n)

        self.lower_std_history = np.zeros(self.history_length_n)
        self.upper_std_history = np.zeros(self.history_length_n)

        self.lower_std_curve = self.velocity_history_plot.plot()
        self.upper_std_curve = self.velocity_history_plot.plot()

        fbi = pg.FillBetweenItem(
            self.lower_std_curve,
            self.upper_std_curve,
            brush=pg.mkBrush(f"{et.utils.color_cycler(0)}50"),
        )

        self.velocity_history_plot.addItem(fbi)

        # PSD plot

        self.psd_plot = self._create_plot(self.plot_layout, row=1, col=0)
        self.psd_plot.setTitle("PSD<br>(colored area represents the slow zone)")
        self.psd_plot.setLabel(axis="left", text="Power")
        self.psd_plot.setLabel(axis="bottom", text="Velocity", units="m/s")
        self.psd_plot.addLegend(labelTextSize="10pt")

        self.psd_smooth_max = et.utils.SmoothMax(tau_grow=0.5, tau_decay=2.0)
        self.psd_curve = self.psd_plot.plot(pen=et.utils.pg_pen_cycler(0), name="PSD")
        self.psd_threshold = self.psd_plot.plot(pen=c0_dashed_pen, name="Threshold")

        psd_slow_zone_color = et.utils.color_cycler(0)
        psd_slow_zone_color = f"{psd_slow_zone_color}50"
        psd_slow_zone_brush = pg.mkBrush(psd_slow_zone_color)

        self.psd_slow_zone = pg.LinearRegionItem(brush=psd_slow_zone_brush, movable=False)
        self.psd_plot.addItem(self.psd_slow_zone)

        brush = et.utils.pg_brush_cycler(0)
        self.psd_peak_plot_item = pg.PlotDataItem(
            pen=None, symbol="o", symbolSize=8, symbolBrush=brush, symbolPen="k"
        )
        self.psd_plot.addItem(self.psd_peak_plot_item)

        self.psd_plot.setLogMode(x=False, y=True)

    def draw_plot_job(self, example_app_result: ExampleAppResult) -> None:
        processor_extra_result = example_app_result.processor_extra_result

        lim = self.velocity_smooth_limits.update(example_app_result.velocity)

        self.velocity_history_plot.setYRange(
            lim[0] - self._VELOCITY_Y_SCALE_MARGIN_M, lim[1] + self._VELOCITY_Y_SCALE_MARGIN_M
        )
        self.velocity_history_plot.setXRange(-self.history_length_s, 0)

        xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.velocity_history = np.roll(self.velocity_history, -1)
        self.velocity_history[-1] = example_app_result.velocity
        self.velocity_curve.setData(xs, self.velocity_history)

        velocity_html = self.psd_html.format(
            f"Distance {np.around(example_app_result.distance_m, 2)} m"
        )
        self.distance_text_item.setHtml(velocity_html)
        self.distance_text_item.setPos(
            -self.history_length_s / 2, lim[1] + self._VELOCITY_Y_SCALE_MARGIN_M
        )

        self.lower_std_history = np.roll(self.lower_std_history, -1)
        self.lower_std_history[-1] = (
            example_app_result.velocity + 0.5 * processor_extra_result.peak_width
        )
        self.lower_std_curve.setData(xs, self.lower_std_history)

        self.upper_std_history = np.roll(self.upper_std_history, -1)
        self.upper_std_history[-1] = (
            example_app_result.velocity - 0.5 * processor_extra_result.peak_width
        )
        self.upper_std_curve.setData(xs, self.upper_std_history)

        low_lim = np.minimum(
            np.min(processor_extra_result.psd_threshold), np.min(processor_extra_result.psd)
        )
        high_lim = self.psd_smooth_max.update(
            np.maximum(processor_extra_result.psd_threshold, processor_extra_result.psd)
        )
        self.psd_plot.setYRange(np.log10(low_lim), np.log10(high_lim))
        self.psd_plot.setXRange(
            processor_extra_result.max_bin_vertical_vs[0],
            processor_extra_result.max_bin_vertical_vs[-1],
        )
        self.psd_curve.setData(
            processor_extra_result.vertical_velocities, processor_extra_result.psd
        )
        self.psd_threshold.setData(
            processor_extra_result.vertical_velocities, processor_extra_result.psd_threshold
        )
        if processor_extra_result.peak_idx is not None:
            self.psd_peak_plot_item.setData(
                [processor_extra_result.vertical_velocities[processor_extra_result.peak_idx]],
                [processor_extra_result.psd[processor_extra_result.peak_idx]],
            )
        else:
            self.psd_peak_plot_item.clear()

        middle_idx = int(np.around(processor_extra_result.vertical_velocities.shape[0] / 2))
        self.psd_slow_zone.setRegion(
            [
                processor_extra_result.vertical_velocities[middle_idx - self.slow_zone],
                processor_extra_result.vertical_velocities[middle_idx + self.slow_zone],
            ]
        )

    @staticmethod
    def _create_plot(parent: pg.GraphicsLayout, row: int, col: int) -> pg.PlotItem:
        velocity_history_plot = parent.addPlot(row=row, col=col)
        velocity_history_plot.setMenuEnabled(False)
        velocity_history_plot.setMouseEnabled(x=False, y=False)
        velocity_history_plot.hideButtons()
        velocity_history_plot.showGrid(x=True, y=True, alpha=0.5)

        return velocity_history_plot


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.range_helper = RangeHelpView()
        scrolly_layout.addWidget(self.range_helper)

        self.config_editor = AttrsConfigEditor(
            title="Example app parameters",
            factory_mapping=self._get_pidget_mapping(),
            config_type=ExampleAppConfig,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "surface_distance": pidgets.FloatPidgetFactory(
                name_label_text="Surface distance:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "surface_distance"),
                suffix=" m",
                decimals=2,
                limits=(0.1, None),
            ),
            "sensor_angle": pidgets.FloatPidgetFactory(
                name_label_text="Sensor angle:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sensor_angle"),
                suffix=" degrees",
                decimals=1,
                limits=(0, 89),
            ),
            "num_points": pidgets.IntPidgetFactory(
                name_label_text="Number of distance points:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "num_points"),
                limits=(1, None),
            ),
            "sweep_rate": pidgets.FloatPidgetFactory(
                name_label_text="Sweep rate:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sweep_rate"),
                suffix=" Hz",
                decimals=1,
                limits=(100, None),
            ),
            "sweeps_per_frame": pidgets.IntPidgetFactory(
                name_label_text="Sweeps per frame:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "sweeps_per_frame"),
                limits=(64, 2048),
            ),
            "hwaas": pidgets.IntPidgetFactory(
                name_label_text="HWAAS:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "hwaas"),
                limits=(1, 511),
            ),
            "profile": pidgets.OptionalEnumPidgetFactory(
                name_label_text="Profile:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "profile"),
                checkbox_label_text="Override",
                enum_type=a121.Profile,
                label_mapping={
                    a121.Profile.PROFILE_1: "1 (shortest)",
                    a121.Profile.PROFILE_2: "2",
                    a121.Profile.PROFILE_3: "3",
                    a121.Profile.PROFILE_4: "4",
                    a121.Profile.PROFILE_5: "5 (longest)",
                },
            ),
            "step_length": pidgets.IntPidgetFactory(
                name_label_text="Step length:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "step_length"),
                limits=(1, None),
            ),
            "frame_rate": pidgets.OptionalFloatPidgetFactory(
                name_label_text="Frame rate:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "frame_rate"),
                checkbox_label_text="Limit",
                suffix=" Hz",
                decimals=1,
                limits=(1, None),
                init_set_value=10,
            ),
            "continuous_sweep_mode": pidgets.CheckboxPidgetFactory(
                name_label_text="Continuos sweep mode",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "continuous_sweep_mode"
                ),
            ),
            "double_buffering": pidgets.CheckboxPidgetFactory(
                name_label_text="Double buffering",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "double_buffering"),
            ),
            "inter_frame_idle_state": pidgets.EnumPidgetFactory(
                enum_type=a121.IdleState,
                name_label_text="Inter frame idle state:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "inter_frame_idle_state"
                ),
                label_mapping={
                    a121.IdleState.DEEP_SLEEP: "Deep sleep",
                    a121.IdleState.SLEEP: "Sleep",
                    a121.IdleState.READY: "Ready",
                },
            ),
            "inter_sweep_idle_state": pidgets.EnumPidgetFactory(
                enum_type=a121.IdleState,
                name_label_text="Inter sweep idle state:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "inter_sweep_idle_state"
                ),
                label_mapping={
                    a121.IdleState.DEEP_SLEEP: "Deep sleep",
                    a121.IdleState.SLEEP: "Sleep",
                    a121.IdleState.READY: "Ready",
                },
            ),
            "time_series_length": pidgets.IntPidgetFactory(
                name_label_text="Time series length:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "time_series_length"),
                limits=(64, None),
            ),
            "psd_lp_coeff": pidgets.FloatSliderPidgetFactory(
                name_label_text="PSD time filtering coeff.:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "psd_lp_coeff"),
                limits=(0, 1),
                decimals=3,
            ),
            "slow_zone": pidgets.IntPidgetFactory(
                name_label_text="Slow zone half length:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "slow_zone"),
                limits=(0, 20),
            ),
            "velocity_lp_coeff": pidgets.FloatSliderPidgetFactory(
                name_label_text="Velocity time filtering coeff.\nper time series:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "velocity_lp_coeff"),
                limits=(0, 1),
                decimals=3,
            ),
            "cfar_win": pidgets.IntPidgetFactory(
                name_label_text="CFAR window length:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "cfar_win"),
                limits=(0, 20),
            ),
            "cfar_guard": pidgets.IntPidgetFactory(
                name_label_text="CFAR guard length:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "cfar_guard"),
                limits=(0, 20),
            ),
            "cfar_sensitivity": pidgets.FloatSliderPidgetFactory(
                name_label_text="Threshold sensitivity:",
                name_label_tooltip=get_attribute_docstring(ExampleAppConfig, "cfar_sensitivity"),
                decimals=2,
                limits=(0.01, 1),
            ),
            "max_peak_interval_s": pidgets.FloatPidgetFactory(
                name_label_text="Max peak interval:",
                name_label_tooltip=get_attribute_docstring(
                    ExampleAppConfig, "max_peak_interval_s"
                ),
                decimals=1,
                limits=(0, 20),
                suffix=" s",
            ),
        }

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
        else:
            self.config_editor.set_data(state.config)
            self.sensor_id_pidget.set_data(state.sensor_id)

            results = state.config._collect_validation_results()

            config_valid = not results

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

            if config_valid:
                self.range_helper.set_data(ExampleApp._get_sensor_config(state.config).subsweep)

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[self.config_editor, self.sensor_id_pidget],
        )

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=self.config_editor.is_ready
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

    def _on_config_update(self, config: ExampleAppConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


SURFACE_VELOCITY_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="surface_velocity",
    title="Surface velocity",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/surface_velocity.html",
    description="Estimate surface speed and direction of streaming water.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional, Sequence


def yn_prompt(message: str) -> bool:
    message = message + " [y/n]"
    inp = input(message)

    while not inp.startswith(("y", "n")):
        print("Please enter 'yes' or 'no':")
        inp = str(input(message))

    return inp.lower().startswith("y")


def number_prompt(message: str) -> int:
    inp = input(message)
    while not inp.isnumeric():
        print("Please enter a digit:")
        inp = input(message)
    return int(inp)


def get_selection_from_user(message: str, choices: Sequence[str]) -> Optional[str]:
    print(message)
    for i, choice in enumerate(choices):
        print(f"{i:<10}{choice}")

    index = number_prompt("Make a selection:")

    try:
        return choices[index]
    except IndexError:
        print(f"Invalid choice: {index}. The valid choices are {list(range(len(choices)))}")
        return None



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._base.calibration import Calibration, CalibrationMapper, _AcceptedFileExtensions, _Path
from ._base.module_info import ModuleFamily, ModuleInfo
from ._standalone_main import main



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from . import hooks
from .pidget_groups import CollapsiblePidgetGroup, FlatPidgetGroup, PidgetGroup, WidgetHook
from .pidgets import (
    WIDGET_WIDTH,
    CheckboxPidget,
    CheckboxPidgetFactory,
    ComboboxPidget,
    ComboboxPidgetFactory,
    EnumPidget,
    EnumPidgetFactory,
    FloatPidget,
    FloatPidgetFactory,
    FloatSliderPidget,
    FloatSliderPidgetFactory,
    IntPidget,
    IntPidgetFactory,
    OptionalEnumPidget,
    OptionalEnumPidgetFactory,
    OptionalFloatPidget,
    OptionalFloatPidgetFactory,
    OptionalIntPidget,
    OptionalIntPidgetFactory,
    OptionalPidget,
    OptionalPidgetFactory,
    Pidget,
    PidgetComboBox,
    PidgetFactory,
    PidgetHook,
    SensorIdPidget,
    SensorIdPidgetFactory,
    StrPidget,
    StrPidgetFactory,
)



# Copyright (c) Acconeer AB, 2024-2025
# All rights reserved

"""
The core of this module is the Node class, which has a type-safe[1]
builder-API to build a "migration tree" from the leaves and up.

An instance of Node can be seen as a node and the node data is a
python type (called 'head').
Each node keeps edges from its children to itself (called 'inbounds').
Children are also Nodes.

Here are some side-by-sides with timeline definitions and the migration tree:

=================================================================================
| Definition                                         | Migration tree           |
=================================================================================
|                                                    |                          |
|                                                    |                          |
| tm.start(int)                                      |  (h=int) <=top           |
|                                                    |                          |
|                                                    |                          |
|----------------------------------------------------|--------------------------|
|                                                    |  (h=int) <=top           |
| tm.start(int)                                      |     ^                    |
|   .load(str, lambda s: int(s), fail=...)           |     | via the lambda     |
|                                                    |     |                    |
|                                                    |  (h=str)                 |
|----------------------------------------------------|--------------------------|
|                                                    |  (h=int) <=top           |
| tm.start(int)                                      |    ^ ^                   |
|   .load(str, lambda s: int(s), fail=...)           |    | |                   |
|   .load(float, lambda s: int(s), fail=...)         |    | +-------+           |
|                                                    |    |         |           |
|                                                    |  (h=str)  (h=float)      |
|----------------------------------------------------|--------------------------|
|  tm.start(int)                                     |  (h=bytes) <=top         |
|    .load(str, lambda s: int(s), fail=...)          |     ^                    |
|    .load(float, lambda s: int(s), fail=...)        |     |                    |
|    .epoch(byte, lambda i: bytes(i), fail=...)      |     |                    |
|                                                    |  (h=int)                 |
|                                                    |    ^ ^                   |
|                                                    |    | |                   |
|                                                    |    | +-------+           |
|                                                    |    |         |           |
|                                                    |  (h=str)  (h=float)      |
|----------------------------------------------------|--------------------------|
|                                                    |  (h=bytes) <=top         |
|  tm.start(int)                                     |     ^ ^                  |
|    .load(str, lambda s: int(s), fail=...)          |     | |                  |
|    .load(float, lambda s: int(s), fail=...         |     | +------+           |
|    .epoch(byte, lambda i: bytes(i), fail=...)      |     |        |           |
|    .load(str,                                      |  (h=int)  (h=str)        |
|          lambda s: bytes(s, encoding="ascii"),     |    ^ ^                   |
|          fail=...)                                 |    | |                   |
|                                                    |    | +-------+           |
|                                                    |    |         |           |
|                                                    |  (h=str)  (h=float)      |
|----------------------------------------------------|--------------------------|

When calling 'migrate', a depth-first-search is done in the
tree to find a matching node for the type of the passed 'obj'.

When a matching Node is found, the "edge" function to the parent node then applied
for each level of the tree until the root is found. If the "edge" function raises
one of the Exceptions specified in 'fail', the search continues in the tree.
If an "edge" function raises and error that is not specified in 'fail', that
exception will be propagated.

If this search fails, a MigrateError is thrown, otherwise a migrated instance of
correct type is returned.

[1] type-safe practically means your editor will yell at you for using the wrong types
"""

from __future__ import annotations

import typing as t

import attrs
import attrs.validators as av
import exceptiongroup as eg
import typing_extensions as te


_T = t.TypeVar("_T")
_R = t.TypeVar("_R")
_S = t.TypeVar("_S")
_P = te.ParamSpec("_P")

_ReqCtxT = t.TypeVar("_ReqCtxT")
_NewCtxT = t.TypeVar("_NewCtxT")
_MigT_contra = t.TypeVar("_MigT_contra", contravariant=True)
_HeadT = t.TypeVar("_HeadT")

Completer: te.TypeAlias = t.Callable[[t.Type[_T]], _T]
_Transform: te.TypeAlias = t.Callable[[_T, Completer[_ReqCtxT]], _R]


def _wrap_with_ignored_completer(f: t.Callable[[_T], _R]) -> _Transform[_T, te.Never, _R]:
    """Turns a regular function into a _Transform for homogenous internal representation.
    Used in the non-contextual 'load' and 'epoch'
    """
    if not callable(f):
        msg = "f must be a callable"
        raise TypeError(msg)

    def wrapper(__x: _T, _: Completer[t.Any]) -> _R:
        return f(__x)

    return wrapper


def _flip(f: t.Callable[[_S, _T], _R]) -> t.Callable[[_T, _S], _R]:
    """Flips the parameters of a 2-parameter function"""

    def wrapper(__x: _T, __y: _S) -> _R:
        return f(__y, __x)

    return wrapper


def _null_completer(typ: type) -> t.Any:
    msg = f"No completion for {typ}"
    raise RuntimeError(msg)


class MigrationError(Exception):
    """This error is raised then a migration fails"""


class MigrationErrorGroup(MigrationError, eg.ExceptionGroup[Exception]):
    """A tree of errors describing why a migration failed"""


def _inbounds_validator(_ignored1: t.Any, _ignored2: t.Any, value: t.Any) -> None:
    (transform, fails, ancestor) = value

    if not callable(transform):
        msg = "The first element of the tuple needs to be a callable"
        raise TypeError(msg)

    if not isinstance(fails, t.Sequence):
        msg = "The second element of the tuple needs to be a sequence"
        raise TypeError(msg)

    if not all(issubclass(x, Exception) for x in fails):
        msg = "All elements of the fail sequence needs to be a subclass of BaseExceptions"
        raise TypeError(msg)

    if not isinstance(ancestor, Node):
        msg = "The third element of the tuple needs to be an Node"
        raise TypeError(msg)


@attrs.frozen(kw_only=True)
class Node(t.Generic[_HeadT, _ReqCtxT, _MigT_contra]):
    """
    A single node in a timeline. The type parameters are

    - _HeadT:       The type this timeline will migrate to.
                    I.e. the most recent version of a class.
    - _ReqCtxT:     The context that might be needed to migrate to '_HeadT'.
                    This is always 'Never' unless 'contextual_load'
                    or 'contextual_epoch' is used.
    - _MigT_contra: The type (usually a 'Union') this timeline is capable of migrating
                    to '_HeadT'.
    """

    head: type[_HeadT] = attrs.field(validator=av.instance_of(type))
    is_epoch: bool
    inbounds: t.Sequence[
        tuple[
            _Transform[t.Any, t.Any, _HeadT],
            t.Sequence[type[Exception]],
            Node[t.Any, t.Any, t.Any],
        ],
    ] = attrs.field(validator=av.deep_iterable(_inbounds_validator))

    def _is_supported(self, typ: type) -> bool:
        return issubclass(typ, self.head) or any(
            ancestor._is_supported(typ) for _, _, ancestor in self.inbounds
        )

    def _node_flatiter(self) -> t.Iterator[Node[t.Any, t.Any, t.Any]]:
        yield self
        for _, _, ancestor in self.inbounds:
            yield from ancestor._node_flatiter()

    def _get_epoch_node_head_types(self) -> set[type]:
        return set(node.head for node in self._node_flatiter() if node.is_epoch)

    def _get_epoch_node_with_head_type(
        self, head_type: type[_T]
    ) -> t.Optional[Node[_T, t.Any, t.Any]]:
        for node in self._node_flatiter():
            if not node.is_epoch:
                continue
            if node.head is head_type:
                return node

        return None

    def load(
        self,
        src: type[_T],
        f: t.Callable[[_T], _HeadT],
        fail: t.Sequence[type[Exception]],
    ) -> Node[_HeadT, _ReqCtxT, _MigT_contra | _T]:
        """Add a loading function to the current epoch"""
        new_gen: Node[_T, te.Never, te.Never] = Node(head=src, is_epoch=False, inbounds=[])
        return Node(
            head=self.head,
            is_epoch=self.is_epoch,
            inbounds=[(_wrap_with_ignored_completer(f), fail, new_gen), *self.inbounds],
        )

    def contextual_load(
        self,
        src: type[_T],
        f: _Transform[_T, _NewCtxT, _HeadT],
        fail: t.Sequence[type[Exception]],
    ) -> Node[_HeadT, _ReqCtxT | _NewCtxT, _MigT_contra | _T]:
        """
        Add a contextual loading function to the current epoch.
        `f`'s second argument should be a `Completer`
        """
        new_gen: Node[_T, te.Never, te.Never] = Node(head=src, is_epoch=False, inbounds=[])
        return Node(
            head=self.head,
            is_epoch=self.is_epoch,
            inbounds=[(f, fail, new_gen), *self.inbounds],
        )

    def epoch(
        self,
        typ: type[_T],
        f: t.Callable[[_HeadT], _T],
        fail: t.Sequence[type[Exception]],
    ) -> Node[_T, _ReqCtxT, _MigT_contra | _HeadT]:
        """Append an epoch to the timeline, adding a level to the migration tree"""
        return Node(
            head=typ,
            is_epoch=True,
            inbounds=[(_wrap_with_ignored_completer(f), fail, self)],
        )

    def contextual_epoch(
        self,
        typ: type[_T],
        f: _Transform[_HeadT, _NewCtxT, _T],
        fail: t.Sequence[type[Exception]],
    ) -> Node[_T, _ReqCtxT | _NewCtxT, _MigT_contra | _HeadT]:
        """Append a contextual epoch to the timeline, adding a level to the migration tree"""
        return Node(
            head=typ,
            is_epoch=True,
            inbounds=[(f, fail, self)],
        )

    def _migrate(
        self,
        obj: _HeadT | _MigT_contra,
        completer: Completer[_ReqCtxT],
    ) -> _HeadT:
        if isinstance(obj, self.head):
            return obj

        if len(self.inbounds) == 0:
            msg = f"Leaf node with head type {self.head} could not migrate object of type {type(obj)}"
            raise MigrationError(msg)

        errors = []

        for transform, transform_fails, ancestor in self.inbounds:
            try:
                ancestor_migrated = ancestor._migrate(obj, completer)
            except Exception as e:
                errors.append(e)
                continue

            try:
                self_migrated = transform(ancestor_migrated, completer)
            except tuple(transform_fails) as e:
                errors.append(e)
            else:
                return self_migrated

        msg = f"Failed migration to {self.head} from {obj!r}"
        raise MigrationErrorGroup(msg, tuple(errors))

    @te.overload
    def migrate(
        self: Node[_HeadT, te.Never, _MigT_contra], obj: _HeadT | _MigT_contra
    ) -> _HeadT: ...

    @te.overload
    def migrate(
        self: Node[_HeadT, te.Never, _MigT_contra],
        obj: _HeadT | _MigT_contra,
        *,
        target_type: type[_T],
    ) -> _T: ...

    @te.overload
    def migrate(
        self: Node[_HeadT, _ReqCtxT, _MigT_contra],
        obj: _HeadT | _MigT_contra,
        *,
        completer: Completer[_ReqCtxT],
    ) -> _HeadT: ...

    @te.overload
    def migrate(
        self: Node[_HeadT, _ReqCtxT, _MigT_contra],
        obj: _HeadT | _MigT_contra,
        *,
        completer: Completer[_ReqCtxT],
        target_type: type[_T],
    ) -> _T: ...

    def migrate(
        self,
        obj: _HeadT | _MigT_contra,
        *,
        completer: Completer[t.Any] = _null_completer,
        target_type: t.Optional[type[_T]] = None,
    ) -> t.Union[_HeadT, _T]:
        """
        Migrate `obj` be the type of the latest epoch or to `target_type` if `target_type` is an "epoch type".

        raises `MigrationError` if migration was not possible
        """
        if target_type is not None:
            target_node = self._get_epoch_node_with_head_type(target_type)

            if target_node is None:
                allowed_target_types = self._get_epoch_node_head_types()
                msg = f"Cannot migrate any object to {target_type}. It's not an epoch type in this tree. 'target_type' is allowed to be one of {allowed_target_types}"
                raise TypeError(msg)

            return target_node.migrate(obj, completer=completer)

        if not self._is_supported(type(obj)):
            msg = f"Cannot migrate objects of type {type(obj)}. It's not part of this tree."
            raise TypeError(msg)

        return self._migrate(obj, completer)

    def nop(self) -> te.Self:
        """Does nothing."""
        return self


def start(typ: type[_T]) -> Node[_T, te.Never, te.Never]:
    return Node(head=typ, is_epoch=True, inbounds=[])



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import typing as t
from enum import IntEnum, auto

from PySide6.QtCore import QSize, Qt, Signal
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QButtonGroup,
    QMainWindow,
    QSizePolicy,
    QSplitter,
    QStackedWidget,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app.new.app_model import AppModel

from .flash_tab import FlashWizard
from .help_tab import HelpMainWidget
from .icons import FLASH, GAUGE, HELP, RECORD
from .misc import ExceptionWidget
from .resource_tab import ResourceMainWidget
from .status_bar import StatusBar
from .stream_tab import StreamingMainWidget
from .utils import LayoutWrapper, TopAlignDecorator


class _IconButton(QToolButton):
    def __init__(
        self,
        icon: QIcon,
        text: str,
        is_active: bool,
        is_checkable: bool,
        icon_size: int = 30,
        tooltip: str = "",
    ) -> None:
        super().__init__()

        self.setCheckable(is_checkable)
        self.setIconSize(QSize(icon_size, icon_size))
        self.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.setSizePolicy(
            QSizePolicy.Policy.Expanding,  # horizontal
            QSizePolicy.Policy.Fixed,  # vertical
        )

        self.setIcon(icon)
        self.setText(text)
        self.setChecked(is_active)
        if tooltip:
            self.setToolTip(tooltip)


class _PagedLayout(QSplitter):
    """
    The widget that is responsible for the paged main widget:

        Buttons          Page
           |              |
           V              V
        +--------+------------------+
        | Stream |                  |
        +--------+                  |
        | Flash  |                  |
        +--------+                  |
        | RC     |                  |
        +--------+                  |
        | Help   |                  |
        +--------+                  |
        |   .    |                  |
        |   .    |                  |
        |        |                  |
        +--------+------------------+
    """

    class _ButtonId(IntEnum):
        STREAM = auto()
        FLASH = auto()
        RESOURCE_CALCULATOR = auto()
        HELP = auto()

    def __init__(
        self,
        app_model: AppModel,
        *,
        index_button_width: int = 60,
    ) -> None:
        super().__init__()

        self.app_model = app_model
        self.app_model.sig_notify.connect(self._on_app_model_update)

        self._index_button_group = QButtonGroup()
        self._index_button_layout = QVBoxLayout()
        self._index_button_layout.setContentsMargins(0, 0, 0, 0)

        self._page_widget = QStackedWidget()

        # Index buttons/Tabs
        stream_index_button = _IconButton(
            RECORD(),
            "Stream",
            is_checkable=True,
            is_active=True,
            tooltip="",
        )
        self._stream_main_widget = StreamingMainWidget(self.app_model, self)
        self._index_button_group.addButton(stream_index_button, id=self._ButtonId.STREAM)
        self._index_button_layout.addWidget(stream_index_button)
        self._page_widget.addWidget(self._stream_main_widget)

        flash_index_button = _IconButton(
            FLASH(),
            "Flash",
            is_checkable=False,
            is_active=False,
            tooltip="",
        )
        self._index_button_group.addButton(flash_index_button, id=self._ButtonId.FLASH)
        self._index_button_layout.addWidget(flash_index_button)

        rc_tooltip = "Resource Calculator"
        rc_index_button = _IconButton(
            GAUGE(),
            "RC",
            is_checkable=True,
            is_active=False,
            tooltip=rc_tooltip,
        )
        self._resource_main_widget = ResourceMainWidget()
        self._index_button_group.addButton(rc_index_button, id=self._ButtonId.RESOURCE_CALCULATOR)
        self._index_button_layout.addWidget(rc_index_button)
        self._page_widget.addWidget(self._resource_main_widget)
        # Signals for "Goto resource calculator".
        self.app_model.sig_resource_tab_input_block_requested.connect(
            self._resource_main_widget.spawn_input_block
        )
        self.app_model.sig_resource_tab_input_block_requested.connect(
            lambda: self.setCurrentWidget(self._resource_main_widget)
        )

        help_index_button = _IconButton(
            HELP(),
            "Help",
            is_checkable=True,
            is_active=False,
            tooltip="",
        )
        self._help_main_widget = HelpMainWidget(self)
        self._index_button_group.addButton(help_index_button, id=self._ButtonId.HELP)
        self._index_button_layout.addWidget(help_index_button)
        self._page_widget.addWidget(self._help_main_widget)

        self._index_button_group.idClicked.connect(self.index_button_clicked)
        self._index_button_widget = TopAlignDecorator(LayoutWrapper(self._index_button_layout))
        self._index_button_widget.setFixedWidth(index_button_width)
        self.addWidget(self._index_button_widget)
        self.addWidget(self._page_widget)
        self.setCollapsible(1, False)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self._index_button_widget.setEnabled(app_model.plugin_state.is_steady)

    def index_button_clicked(self, index_button_id: int) -> None:
        if index_button_id == self._ButtonId.STREAM:
            self.setCurrentWidget(self._stream_main_widget)
        elif index_button_id == self._ButtonId.FLASH:
            FlashWizard(self.app_model).exec()
        elif index_button_id == self._ButtonId.RESOURCE_CALCULATOR:
            self.setCurrentWidget(self._resource_main_widget)
        elif index_button_id == self._ButtonId.HELP:
            self.setCurrentWidget(self._help_main_widget)
        else:
            msg = f"Unknown index_button_id: {index_button_id}"
            raise ValueError(msg)

    def setCurrentWidget(self, widget: QWidget) -> None:
        if widget is self._stream_main_widget:
            self._page_widget.setCurrentWidget(self._stream_main_widget)
            self._index_button_group.button(self._ButtonId.STREAM).setChecked(True)
        elif widget is self._resource_main_widget:
            self._page_widget.setCurrentWidget(self._resource_main_widget)
            self._index_button_group.button(self._ButtonId.RESOURCE_CALCULATOR).setChecked(True)
        elif widget is self._help_main_widget:
            self._page_widget.setCurrentWidget(self._help_main_widget)
            self._index_button_group.button(self._ButtonId.HELP).setChecked(True)
        else:
            msg = f"Unknown widget: {widget}"
            raise ValueError(msg)


class MainWindow(QMainWindow):
    sig_closing = Signal()

    def __init__(self, app_model: AppModel) -> None:
        super().__init__()

        self.resize(1280, 720)

        paged_layout = _PagedLayout(app_model)

        self.setCentralWidget(paged_layout)

        self.setStatusBar(StatusBar(app_model, self))
        self.setWindowTitle("Acconeer Exploration Tool")
        self.moveEvent = lambda _: self.saveGeometry()  # type: ignore[method-assign,assignment]

        app_model.sig_error.connect(self.on_app_model_error)

    def on_app_model_error(self, exception: Exception, traceback_str: t.Optional[str]) -> None:
        ExceptionWidget(self, exc=exception, traceback_str=traceback_str).exec()

    def closeEvent(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.sig_closing.emit()
        return super().closeEvent(*args, **kwargs)
